import importlib
import inspect
import pkgutil
import warnings
from pathlib import Path
from typing import Union, List

from pymodaq.utils.gui_utils.dock import DockArea
from pymodaq.utils.daq_utils import get_plugins
from pymodaq.utils.logger import get_module_name, set_logger
from pymodaq.utils.daq_utils import find_dict_in_list_from_key_val, get_entrypoints
from pymodaq.utils.data import DataToExport, DataCalculated, DataActuator

logger = set_logger(get_module_name(__file__))

DAQ_Move_Stage_type = get_plugins('daq_move')
DAQ_0DViewer_Det_types = get_plugins('daq_0Dviewer')
DAQ_1DViewer_Det_types = get_plugins('daq_1Dviewer')
DAQ_2DViewer_Det_types = get_plugins('daq_2Dviewer')
DAQ_NDViewer_Det_types = get_plugins('daq_NDviewer')


class DataToActuatorPID(DataToExport):
    """ Particular case of a DataToExport adding one named parameter to indicate what kind of change should be applied
    to the actuators, absolute or relative

    Attributes
    ----------
    mode: str
        Adds an attribute called mode holding a string describing the type of change: relative or absolute

    Parameters
    ---------
    mode: str
        either 'rel' or 'abs' for a relative or absolute change of the actuator's values
    """

    def __init__(self, *args, mode='rel', **kwargs):
        if mode not in ['rel', 'abs']:
            warnings.warn('Incorrect mode for the actuators, switching to default relative mode: rel')
            mode = 'rel'
        kwargs.update({'mode': mode})
        super().__init__(*args, **kwargs)

    def __repr__(self):
        return f'{super().__repr__()}: {self.mode}'


class PIDModelGeneric:
    limits = dict(max=dict(state=False, value=1),
                  min=dict(state=False, value=0),)
    konstants = dict(kp=1, ki=0.1, kd=0.001)
    params = []

    Nsetpoints = 1
    setpoint_ini = [0. for ind in range(Nsetpoints)]
    setpoints_names = ['' for ind in range(Nsetpoints)]

    actuators_name = []
    detectors_name = []

    epsilon = 1

    def __init__(self, pid_controller):
        self.pid_controller = pid_controller  # instance of the pid_controller using this model
        self.modules_manager = pid_controller.modules_manager

        self.settings = self.pid_controller.settings.child('models', 'model_params')  # set of parameters
        self.data_names = None
        self.curr_output = [0. for ind in range(self.Nsetpoints)]
        self.curr_input = None

        self.check_modules(self.modules_manager)

    def setpoint(self, values):
        self.pid_controller.setpoints = values

    def apply_constants(self):
        for kxx in self.konstants:
            self.pid_controller.settings.child('main_settings', 'pid_controls', 'pid_constants',
                                               kxx).setValue(self.konstants[kxx])

    def apply_limits(self):
        for limit in self.limits:
            self.pid_controller.settings.child('main_settings', 'pid_controls', 'output_limits',
                                               f'output_limit_{limit}').setValue(self.limits[limit]['value'])
            self.pid_controller.settings.child('main_settings', 'pid_controls', 'output_limits',
                                               f'output_limit_{limit}_enabled').setValue(self.limits[limit]['state'])

    def check_modules(self, modules_manager):
        for act in self.actuators_name:
            if act not in modules_manager.actuators_name:
                logger.warning(f'The actuator {act} defined in the PID model is not present in the Dashboard')
                return False
        for det in self.detectors_name:
            if det not in modules_manager.detectors_name:
                logger.warning(f'The detector {det} defined in the PID model is not present in the Dashboard')

    def update_detector_names(self):
        names = self.pid_controller.settings.child('main_settings', 'detector_modules').value()['selected']
        self.data_names = []
        for name in names:
            name = name.split('//')
            self.data_names.append(name)

    def update_settings(self, param):
        """
        Get a parameter instance whose value has been modified by a user on the UI
        To be overwritten in child class
        """
        if param.name() == '':
            pass

    def ini_model(self):
        self.apply_limits()
        self.setpoint(self.setpoint_ini)
        self.apply_constants()

    def convert_input(self, measurements: DataToExport) -> DataToExport:
        """
        Convert the measurements in the units to be fed to the PID (same dimensionality as the setpoint)
        Parameters
        ----------
        measurements: DataToExport
         DataToExport object from which the model extract a value of the same units as the setpoint

        Returns
        -------
        DataToExport: the converted input as 0D DataCalculated stored in a DataToExport
        """
        raise NotImplementedError

    def convert_output(self, outputs: List[float], dt, stab=True) -> DataToActuatorPID:
        """
        Convert the output of the PID in units to be fed into the actuator
        Parameters
        ----------
        outputs: (list of float) output value from the PID from which the model extract a value of the same units as the actuator
        dt: (float) elapsed time in seconds since last call
        Returns
        -------
        DataToActuatorPID: the converted output as a DataToActuatorPID object (derived from DataToExport)

        """
        self.curr_output = outputs
        return DataToActuatorPID('pid', mode='rel',
                                 data=[DataActuator(self.actuators_name[ind], data=outputs[ind])
                                       for ind in range(len(outputs))])


def main(xmlfile):
    from pymodaq.dashboard import DashBoard
    from pymodaq.utils.config import get_set_preset_path
    from pathlib import Path
    from qtpy import QtWidgets

    import sys
    app = QtWidgets.QApplication(sys.argv)
    win = QtWidgets.QMainWindow()
    area = DockArea()
    win.setCentralWidget(area)
    win.resize(1000, 500)
    win.setWindowTitle('PyMoDAQ Dashboard')

    dashboard = DashBoard(area)
    file = Path(get_set_preset_path()).joinpath(xmlfile)
    if file.exists():
        dashboard.set_preset_mode(file)
        # prog.load_scan_module()
        pid_area = DockArea()
        pid_window = QtWidgets.QMainWindow()
        pid_window.setCentralWidget(pid_area)

        prog = dashboard.load_pid_module(pid_window)

    else:
        msgBox = QtWidgets.QMessageBox()
        msgBox.setText(f"The default file specified in the configuration file does not exists!\n"
                       f"{file}\n"
                       f"Impossible to load the DAQ_PID Module")
        msgBox.setStandardButtons(msgBox.Ok)
        ret = msgBox.exec()

    sys.exit(app.exec_())


def get_models(model_name=None):
    """
    Get PID Models as a list to instantiate Control Actuators per degree of liberty in the model

    Returns
    -------
    list: list of disct containting the name and python module of the found models
    """
    from pymodaq.extensions.pid.utils import PIDModelGeneric
    models_import = []
    discovered_models = get_entrypoints(group='pymodaq.pid_models')
    if len(discovered_models) > 0:
        for pkg in discovered_models:
            try:
                module = importlib.import_module(pkg.value)
                module_name = pkg.value

                for mod in pkgutil.iter_modules([str(Path(module.__file__).parent.joinpath('models'))]):
                    try:
                        model_module = importlib.import_module(f'{module_name}.models.{mod.name}', module)
                        classes = inspect.getmembers(model_module, inspect.isclass)
                        for name, klass in classes:
                            if klass.__base__ is PIDModelGeneric:
                                models_import.append({'name': mod.name, 'module': model_module, 'class': klass})
                                break

                    except Exception as e:  # pragma: no cover
                        logger.warning(str(e))

            except Exception as e:  # pragma: no cover
                logger.warning(f'Impossible to import the {pkg.value} extension: {str(e)}')

    if model_name is None:
        return models_import
    else:
        return find_dict_in_list_from_key_val(models_import, 'name', model_name)



# -*- coding: utf-8 -*-
"""
Created the 19/11/2023

@author: Sebastien Weber
"""

from pathlib import Path
from pymodaq.utils.config import BaseConfig, getitem_recursive


class ScanConfig(BaseConfig):
    """Main class to deal with configuration values for this plugin"""
    config_template_path = Path(__file__).parent.parent.parent.joinpath('resources/config_scan_template.toml')
    config_name = f"scanner_settings"

    def __setitem__(self, key, value):
        if isinstance(key, tuple):
            dic = getitem_recursive(self._config, *key, ndepth=1, create_if_missing=True)
            if value is None:  # means the setting is a group
                value = {}
            dic[key[-1]] = value
        else:
            self._config[key] = value


from qtpy import QtWidgets
from qtpy.QtCore import Qt, QObject

import pymodaq.utils.parameter.ioxml
from pymodaq.utils.parameter import ParameterTree, Parameter

import sys
from pymodaq.utils.plotting.data_viewers.viewer2D import Viewer2D
from pymodaq.post_treatment.daq_measurement.daq_measurement_main import DAQ_Measurement
from pymodaq.utils.h5modules.browsing import browse_data
import os
from pymodaq.utils.config import get_set_local_dir

local_path = get_set_local_dir()
calib_path = os.path.join(local_path, 'camera_calibrations')
if not os.path.isdir(calib_path):
    os.makedirs(calib_path)


class CalibrationCamera(QtWidgets.QWidget, QObject):
    def __init__(self, parent=None, h5filepath=None):

        
        super(CalibrationCamera, self).__init__()
        if parent is None:
            parent = QtWidgets.QWidget()

        self.parent = parent
        self.setupUI()
        self.fname = None
        self.node_path = None

        self.viewer2D.get_action('histo').trigger()
        self.viewer2D.get_action('roi').trigger()

        self.meas_module.ui.measurement_type_combo.setCurrentText('Sinus')
        QtWidgets.QApplication.processEvents()
        self.meas_module.ui.measure_subtype_combo.setCurrentText('dx')

    def setupUI(self):

        params = [{'title': 'Load data:', 'name': 'Load data', 'type': 'action', },
                  {'title': 'Set Measurement:', 'name': 'Do measurement', 'type': 'action', },
                  {'title': 'Calib from:', 'name': 'calib_from', 'type': 'list', 'limits': ['Hlineout', 'Vlineout'], },
                  {'title': 'X axis:', 'name': 'xaxis', 'type': 'group', 'children': [
                      {'title': 'Units:', 'name': 'xunits', 'type': 'str', 'value': "µm"},
                      {'title': 'dx (units):', 'name': 'dx_units', 'type': 'float', 'value': 0, },
                      {'title': 'dx from fit:', 'name': 'dx_fit', 'type': 'float', 'value': 0, },
                      {'title': 'x Offset:', 'name': 'xoffset', 'type': 'float', 'value': 0., 'readonly': False}, ]},
                  {'title': 'Y axis:', 'name': 'yaxis', 'type': 'group', 'children': [
                      {'title': 'Units:', 'name': 'yunits', 'type': 'str', 'value': "µm"},
                      {'title': 'dy (units):', 'name': 'dy_units', 'type': 'float', 'value': 0., },
                      {'title': 'dy from fit:', 'name': 'dy_fit', 'type': 'float', 'value': 0., },
                      {'title': 'y Offset:', 'name': 'yoffset', 'type': 'float', 'value': 0., 'readonly': False}, ]},
                  {'title': 'Save calib:', 'name': 'Save data', 'type': 'action', },
                  {'title': 'Data Saved:', 'name': 'data_saved', 'type': 'led', 'value': False, 'readonly': True},
                  {'title': 'File path:', 'name': 'filepath', 'type': 'text', 'value': '', 'readonly': True},
                  ]

        layout = QtWidgets.QHBoxLayout()
        self.parent.setLayout(layout)

        splitter_H = QtWidgets.QSplitter(Qt.Horizontal)

        splitter = QtWidgets.QSplitter(Qt.Vertical)

        widg2D = QtWidgets.QWidget()
        self.viewer2D = Viewer2D(widg2D)
        self.viewer2D.title = 'calib'
        self.viewer2D.get_action('autolevels').trigger()

        splitter.addWidget(widg2D)

        widg_meas = QtWidgets.QWidget()
        self.meas_module = DAQ_Measurement(widg_meas)
        splitter.addWidget(widg_meas)

        self.settings_tree = ParameterTree()
        self.settings = Parameter.create(name='Settings', type='group', children=params)
        self.settings_tree.setParameters(self.settings, showTop=False)
        self.settings.sigTreeStateChanged.connect(self.parameter_tree_changed)

        self.settings.child(('Load data')).sigActivated.connect(self.load_data)
        self.settings.child(('Save data')).sigActivated.connect(self.save_data)
        self.settings.child(('Do measurement')).sigActivated.connect(self.update_data_meas)

        splitter_H.addWidget(self.settings_tree)
        splitter_H.addWidget(splitter)

        layout.addWidget(splitter_H)

    def update_data_meas(self):
        if self.settings.child(('calib_from')).value() == 'Hlineout':
            xdata = self.viewer2D.data_to_export['data1D']['calib_Hlineout_ROI_00']['x_axis']
            ydata = self.viewer2D.data_to_export['data1D']['calib_Hlineout_ROI_00']['data']
        else:
            xdata = self.viewer2D.data_to_export['data1D']['calib_Vlineout_ROI_00']['x_axis']
            ydata = self.viewer2D.data_to_export['data1D']['calib_Vlineout_ROI_00']['data']

        self.meas_module.update_data(xdata, ydata)

    def load_data(self):
        self.data, self.fname, self.node_path = browse_data(ret_all=True)
        if self.data is not None:
            self.settings.child('xaxis', 'xoffset').setValue((-self.data.shape[1] / 2))
            self.settings.child('yaxis', 'yoffset').setValue((-self.data.shape[0] / 2))
            self.viewer2D.setImage(self.data)
            self.viewer2D.get_action('histo').trigger()
            self.viewer2D.get_action('roi').trigger()
            QtWidgets.QApplication.processEvents()

    def save_data(self):
        params = [{'title': 'Axis options:', 'name': 'axes', 'type': 'group', 'visible': False, 'expanded': False,
                   'children': [
                       {'title': 'X axis:', 'name': 'xaxis', 'type': 'group', 'children': [
                           {'title': 'Label:', 'name': 'xlabel', 'type': 'str', 'value': "x axis"},
                           {'title': 'Units:', 'name': 'xunits', 'type': 'str',
                            'value': self.settings.child('xaxis', 'xunits').value()},
                           {'title': 'Offset:', 'name': 'xoffset', 'type': 'float', 'default': 0.,
                            'value': self.settings.child('xaxis', 'xoffset').value()},

                           {'title': 'Scaling', 'name': 'xscaling', 'type': 'float', 'default': 1.,
                            'value': self.settings.child('xaxis', 'dx_units').value() / self.settings.child('xaxis',
                                                                                                            'dx_fit').value()},
                       ]},
                       {'title': 'Y axis:', 'name': 'yaxis', 'type': 'group', 'children': [
                           {'title': 'Label:', 'name': 'ylabel', 'type': 'str', 'value': "x axis"},
                           {'title': 'Units:', 'name': 'yunits', 'type': 'str',
                            'value': self.settings.child('yaxis', 'yunits').value()},
                           {'title': 'Offset:', 'name': 'yoffset', 'type': 'float', 'default': 0.,
                            'value': self.settings.child('yaxis', 'yoffset').value()},
                           {'title': 'Scaling', 'name': 'yscaling', 'type': 'float', 'default': 1.,
                            'value': self.settings.child('yaxis', 'dy_units').value() / self.settings.child('yaxis',
                                                                                                            'dy_fit').value()},
                       ]},
                   ]}, ]
        param_obj = Parameter.create(name='Axes_Settings', type='group', children=params)
        pymodaq.utils.parameter.ioxml.parameter_to_xml_file(param_obj,
                                                                os.path.join(calib_path, os.path.split(self.fname)[1]))
        self.settings.child(('data_saved')).setValue(True)
        self.settings.child(('filepath')).setValue(os.path.join(calib_path, os.path.split(self.fname)[1]))

    def parameter_tree_changed(self, param, changes):
        """

        """
        for param, change, data in changes:
            path = self.settings.childPath(param)
            if path is not None:
                childName = '.'.join(path)
            else:
                childName = param.name()
            if change == 'childAdded':
                pass

            elif change == 'value':
                if param.name() == 'calib_from':
                    self.update_data_meas()
                elif param.name() == 'dx_fit':
                    scaling = self.settings.child('xaxis', 'dx_units').value() / param.value()
                    self.settings.child('xaxis', 'xoffset').setValue(-scaling * self.data.shape[1] / 2)
                elif param.name() == 'dy_fit':
                    scaling = self.settings.child('yaxis', 'dy_units').value() / param.value()
                    self.settings.child('yaxis', 'yoffset').setValue(-scaling * self.data.shape[0] / 2)

                self.settings.child(('data_saved')).setValue(False)

            elif change == 'parent':
                pass


if __name__ == '__main__':
    app = QtWidgets.QApplication(sys.argv)
    win = QtWidgets.QWidget()
    prog = CalibrationCamera(win)
    win.show()
    sys.exit(app.exec_())



# -*- coding: utf-8 -*-
"""
Created the 06/12/2022

@author: Sebastien Weber
"""
import sys
from typing import List, TYPE_CHECKING

from qtpy import QtWidgets, QtCore
from qtpy.QtCore import Signal

from pymodaq.utils.gui_utils import CustomApp
from pymodaq.utils.logger import set_logger, get_module_name
from pymodaq.utils.config import Config, get_set_preset_path
from pymodaq.utils.gui_utils import DockArea, Dock
from pymodaq.utils.gui_utils.widgets.spinbox import QSpinBox_ro
from pymodaq.utils.parameter.pymodaq_ptypes.led import QLED
from pymodaq.utils.plotting.data_viewers.viewer import ViewerDispatcher, ViewersEnum
from pymodaq.utils.daq_utils import ThreadCommand

if TYPE_CHECKING:
    from pymodaq.utils.parameter import ParameterTree

config = Config()
logger = set_logger(get_module_name(__file__))


class DAQScanUI(CustomApp, ViewerDispatcher):
    """

    """
    command_sig = Signal(ThreadCommand)

    def __init__(self, parent):
        CustomApp.__init__(self, parent)
        self.setup_docks()
        ViewerDispatcher.__init__(self, self.dockarea, title='Scanner', next_to_dock=self.dock_command)

        self.setup_menu()
        self.setup_actions()
        self.connect_things()

    def setup_actions(self):
        self.add_action('quit', 'Quit the module', 'close2', menu=self.file_menu)
        self.add_action('ini_positions', 'Init Positions', '', menu=self.action_menu)
        self.add_action('start', 'Start Scan', 'run2', "Start the scan", menu=self.action_menu)
        self.add_action('start_batch', 'Start ScanBatches', 'run_all', "Start the batch of scans", menu=self.action_menu)
        self.add_action('stop', 'Stop Scan', 'stop', "Stop the scan", menu=self.action_menu)
        self.add_action('move_at', 'Move at doubleClicked', 'move_contour',
                        "Move to positions where you double clicked", checkable=True, menu=self.action_menu)
        self.add_action('log', 'Show Log file', 'information2', menu=self.file_menu)

        self.add_action('load', 'Load File', 'Open', menu=self.file_menu, auto_toolbar=False)
        self.file_menu.addSeparator()
        self.add_action('save', 'Save file as', 'SaveAs', menu=self.file_menu, auto_toolbar=False)
        self.add_action('show_file', 'Show file content', '', menu=self.file_menu, auto_toolbar=False)

        self.add_action('navigator', 'Show Navigator', '', menu=self._extensions_menu, auto_toolbar=False)
        self.add_action('batch', 'Show Batch Scanner', '', menu=self._extensions_menu, auto_toolbar=False)
        self.set_action_visible('start_batch', False)

    def enable_start_stop(self, enable=True):
        """If True enable main buttons to launch/stop scan"""
        self.set_action_enabled('start', enable)
        self.set_action_enabled('stop', enable)

    def connect_things(self):
        self.connect_action('quit', lambda: self.command_sig.emit(ThreadCommand('quit')))
        self.connect_action('ini_positions', lambda: self.command_sig.emit(ThreadCommand('ini_positions')))
        self.connect_action('start', lambda: self.command_sig.emit(ThreadCommand('start')))
        self.connect_action('start_batch', lambda: self.command_sig.emit(ThreadCommand('start_batch')))
        self.connect_action('stop', lambda: self.command_sig.emit(ThreadCommand('stop')))
        self.connect_action('move_at', lambda: self.command_sig.emit(ThreadCommand('move_at')))
        self.connect_action('log', lambda: self.command_sig.emit(ThreadCommand('show_log', )))

        self.connect_action('load', lambda: self.command_sig.emit(ThreadCommand('load')))
        self.connect_action('save', lambda: self.command_sig.emit(ThreadCommand('save')))
        self.connect_action('show_file', lambda: self.command_sig.emit(ThreadCommand('show_file')))
        self.connect_action('navigator', lambda: self.command_sig.emit(ThreadCommand('navigator')))
        self.connect_action('batch', lambda: self.command_sig.emit(ThreadCommand('batch')))

    def setup_menu(self):
        self.menubar = QtWidgets.QMenuBar()
        self.menubar.setMaximumHeight(30)
        self.mainwindow.setMenuBar(self.menubar)
        self.file_menu = self.menubar.addMenu('File')
        self._extensions_menu = self.menubar.addMenu('Extensions')
        self.action_menu = self.menubar.addMenu('Actions')

    def setup_docks(self):
        self.dock_command = Dock('Scan Command')
        self.dockarea.addDock(self.dock_command)

        widget_command = QtWidgets.QWidget()
        widget_command.setLayout(QtWidgets.QVBoxLayout())
        self.dock_command.addWidget(widget_command)
        widget_command.layout().addWidget(self._toolbar)

        splitter_widget = QtWidgets.QSplitter(QtCore.Qt.Horizontal)
        splitter_v_widget = QtWidgets.QSplitter(QtCore.Qt.Vertical)
        widget_command.layout().addWidget(splitter_widget)
        splitter_widget.addWidget(splitter_v_widget)
        self.module_widget = QtWidgets.QWidget()
        self.module_widget.setLayout(QtWidgets.QVBoxLayout())
        self.module_widget.setMinimumWidth(220)
        self.module_widget.setMaximumWidth(400)

        self.plotting_widget = QtWidgets.QWidget()
        self.plotting_widget.setLayout(QtWidgets.QVBoxLayout())
        self.plotting_widget.setMinimumWidth(220)
        self.plotting_widget.setMaximumWidth(400)

        settings_widget = QtWidgets.QWidget()
        settings_widget.setLayout(QtWidgets.QVBoxLayout())
        settings_widget.setMinimumWidth(220)

        splitter_v_widget.addWidget(self.module_widget)
        splitter_v_widget.addWidget(self.plotting_widget)

        splitter_v_widget.setSizes([400, 400])
        splitter_widget.addWidget(settings_widget)


        self._statusbar = QtWidgets.QStatusBar()
        self.mainwindow.setStatusBar(self._statusbar)
        self.populate_status_bar()

        self.settings_toolbox = QtWidgets.QToolBox()
        settings_widget.layout().addWidget(self.settings_toolbox)
        self.scanner_widget = QtWidgets.QWidget()
        self.scanner_widget.setLayout(QtWidgets.QVBoxLayout())
        self.settings_toolbox.addItem(self.scanner_widget, 'Scanner Settings')

    def add_settings_toolbox_widget(self, widget: QtWidgets.QWidget, name: str):
        """Add a widget, usaually a ParameterTree to the SettingsToolbox"""
        self.settings_toolbox.addItem(widget, name)

    def add_scanner_settings(self, tree: 'ParameterTree'):
        """Adds a  ParameterTree to the Scanner settings widget"""
        self.scanner_widget.layout().addWidget(tree)

    def populate_toolbox_widget(self, widgets: List[QtWidgets.QWidget], names: List[str]):
        for widget, name in zip(widgets, names):
            self.settings_toolbox.addItem(widget, name)

    def set_scanner_settings(self, settings_tree: QtWidgets.QWidget):
        while True:
            child = self.scanner_widget.layout().takeAt(0)
            if not child:
                break
            child.widget().deleteLater()
            QtWidgets.QApplication.processEvents()

        self.scanner_widget.layout().addWidget(settings_tree)

    def set_modules_settings(self, settings_widget):
        self.module_widget.layout().addWidget(settings_widget)

    def set_plotting_settings(self, settings_plotting):
        self.plotting_widget.layout().addWidget(settings_plotting)

    def populate_status_bar(self):
        self._status_message_label = QtWidgets.QLabel('Initializing')
        self._n_scan_steps_sb = QSpinBox_ro()
        self._n_scan_steps_sb.setToolTip('Total number of steps')
        self._indice_scan_sb = QSpinBox_ro()
        self._indice_scan_sb.setToolTip('Current step value')
        self._indice_average_sb = QSpinBox_ro()
        self._indice_average_sb.setToolTip('Current average value')
        
        self._scan_done_LED = QLED()
        self._scan_done_LED.set_as_false()
        self._scan_done_LED.clickable = False
        self._scan_done_LED.setToolTip('Scan done state')
        self._statusbar.addPermanentWidget(self._status_message_label)

        self._statusbar.addPermanentWidget(self._n_scan_steps_sb)
        self._statusbar.addPermanentWidget(self._indice_scan_sb)
        self._statusbar.addPermanentWidget(self._indice_average_sb)
        self._indice_average_sb.setVisible(False)
        self._statusbar.addPermanentWidget(self._scan_done_LED)

    @property
    def n_scan_steps(self):
        return self._n_scan_steps_sb.value()

    @n_scan_steps.setter
    def n_scan_steps(self, nsteps: int):
        self._n_scan_steps_sb.setValue(nsteps)

    def display_status(self, status: str, wait_time=1000):
        self._statusbar.showMessage(status, wait_time)
        
    def set_permanent_status(self, status: str):
        self._status_message_label.setText(status)

    def set_scan_step(self, step_ind: int):
        self._indice_scan_sb.setValue(step_ind)

    def show_average_step(self, show: bool = True):
        self._indice_average_sb.setVisible(show)

    def set_scan_step_average(self, step_ind: int):
        self._indice_average_sb.setValue(step_ind)

    def set_scan_done(self, done=True):
        self._scan_done_LED.set_as(done)

    def update_viewers(self, viewers_type: List[ViewersEnum], viewers_name: List[str] = None, force=False):
        super().update_viewers(viewers_type, viewers_name, force)
        self.command_sig.emit(ThreadCommand('viewers_changed', attribute=dict(viewer_types=self.viewer_types,
                                                                              viewers=self.viewers)))

def main():

    app = QtWidgets.QApplication(sys.argv)

    win = QtWidgets.QMainWindow()
    dockarea = DockArea()
    win.setCentralWidget(dockarea)
    win.resize(1000, 500)
    win.setWindowTitle('DAQScan Extension')

    prog = DAQScanUI(dockarea)
    win.show()


    def print_command_sig(cmd_sig):
        print(cmd_sig)
        prog.display_status(str(cmd_sig))

    prog.command_sig.connect(print_command_sig)
    prog.update_viewers([ViewersEnum['Viewer0D'], ViewersEnum['Viewer1D'], ViewersEnum['Viewer2D']])

    sys.exit(app.exec_())


if __name__ == '__main__':
    main()



from qtpy import QtWidgets
from qtpy.QtCore import Qt, QObject, Slot, Signal

import sys
from pymodaq.post_treatment.daq_measurement.daq_measurement_GUI import Ui_Form
from pymodaq.utils import daq_utils as utils
from pymodaq.utils.plotting.utils.filter import FourierFilterer
from scipy.optimize import curve_fit
import numpy as np
from pymodaq.utils.enums import BaseEnum


class Measurement_type(BaseEnum):
    Cursor_Integration = 0
    Max = 1
    Min = 2
    Gaussian_Fit = 3
    Lorentzian_Fit = 4
    Exponential_Decay_Fit = 5
    Sinus = 6

    @classmethod
    def update_measurement_subtype(cls, mtype):
        measurement_gaussian_subitems = ["amp", "dx", "x0", "offset"]
        measurement_laurentzian_subitems = ["alpha", "gamma", "x0", "offset"]
        measurement_decay_subitems = ["N0", "gamma", 'x0', "offset"]
        measurement_cursor_subitems = ["sum", "mean", "std"]
        measurement_sinus_subtitems = ['A', 'dx', 'phi', 'offset']
        variables = ", "
        formula = ""
        subitems = []
        if mtype == 'Cursor_Integration':  # "Cursor integration":
            subitems = measurement_cursor_subitems

        if mtype == 'Gaussian_Fit':  # "Gaussian Fit":
            subitems = measurement_gaussian_subitems
            formula = "amp*np.exp(-2*np.log(2)*(x-x0)**2/dx**2)+offset"
            variables = variables.join(measurement_gaussian_subitems)

        elif mtype == 'Lorentzian_Fit':  # "Lorentzian Fit":
            subitems = measurement_laurentzian_subitems
            variables = variables.join(measurement_laurentzian_subitems)
            formula = "alpha/np.pi*gamma/2/((x-x0)**2+(gamma/2)**2)+offset"

        elif mtype == 'Exponential_Decay_Fit':  # "Exponential Decay Fit":
            subitems = measurement_decay_subitems
            variables = variables.join(measurement_decay_subitems)
            formula = "N0*np.exp(-(x-x0)/gamma)+offset"

        elif mtype == 'Sinus':
            subitems = measurement_sinus_subtitems
            variables = variables.join(measurement_sinus_subtitems)
            formula = "A*np.sin(2*np.pi*x/dx-phi)+offset"

        return [variables, formula, subitems]

    def gaussian_func(self, x, amp, dx, x0, offset):
        return amp * np.exp(-2 * np.log(2) * (x - x0) ** 2 / dx ** 2) + offset

    def laurentzian_func(self, x, alpha, gamma, x0, offset):
        return alpha / np.pi * 1 / 2 * gamma / ((x - x0) ** 2 + (1 / 2 * gamma) ** 2) + offset

    def decaying_func(self, x, N0, gamma, x0, offset):
        return N0 * np.exp(-(x - x0) / gamma) + offset

    def sinus_func(self, x, A, dx, phi, offset):
        return A * np.sin(2 * np.pi * x / dx - phi) + offset


class DAQ_Measurement(Ui_Form, QObject):
    """
        =================== ================================== =======================================
        **Attributes**       **Type**                          **Description**


        *ui*                 QObject                           The local instance of User Interface
        *wait_time*          int                               The default delay of showing
        *parent*             QObject                           The QObject initializing the UI
        *xdata*              1D numpy array                    The x axis data
        *ydata*              1D numpy array                    The y axis data
        *measurement_types*  instance of daq_utils.DAQ_enums   The type of the measurement, between:
                                                                    * 'Cursor_Integration'
                                                                    * 'Max'
                                                                    * 'Min'
                                                                    * 'Gaussian_Fit'
                                                                    * 'Lorentzian_Fit'
                                                                    * 'Exponential_Decay_Fit'
        =================== ================================== =======================================

        References
        ----------
        Ui_Form, QObject, qtpy, pyqtgraph
    """
    measurement_signal = Signal(list)

    def __init__(self, parent):
        
        super(Ui_Form, self).__init__()
        self.ui = Ui_Form()
        self.ui.setupUi(parent)

        self.widg = QtWidgets.QWidget()
        self.fourierfilt = FourierFilterer(self.widg)
        self.ui.splitter_2.addWidget(self.widg)
        self.ui.graph1D = self.fourierfilt.viewer1D.plotwidget
        QtWidgets.QApplication.processEvents()

        self.ui.splitter.setSizes([200, 400])
        self.ui.statusbar = QtWidgets.QStatusBar(parent)
        self.ui.statusbar.setMaximumHeight(15)
        self.ui.StatusBarLayout.addWidget(self.ui.statusbar)
        self.wait_time = 1000
        self.parent = parent
        self.xdata = None
        self.ydata = None

        self.measurement_types = Measurement_type.names()
        self.measurement_type = Measurement_type(0)
        self.ui.measurement_type_combo.clear()
        self.ui.measurement_type_combo.addItems(self.measurement_types)

        self.ui.fit_curve = self.fourierfilt.viewer1D.plotwidget.plot()
        self.ui.fit_curve.setPen("y")
        self.ui.fit_curve.setVisible(False)

        self.ui.selected_region = self.fourierfilt.viewer1D.ROI
        self.ui.selected_region.setZValue(-10)
        self.ui.selected_region.setBrush('b')
        self.ui.selected_region.setOpacity(0.2)
        self.ui.selected_region.setVisible(True)
        self.fourierfilt.viewer1D.plotwidget.addItem(self.ui.selected_region)

        # Connecting buttons:
        self.ui.Quit_pb.clicked.connect(self.Quit_fun, type=Qt.QueuedConnection)
        self.ui.measurement_type_combo.currentTextChanged[str].connect(self.update_measurement_subtype)
        self.ui.measure_subtype_combo.currentTextChanged[str].connect(self.update_measurement)
        self.update_measurement_subtype(self.ui.measurement_type_combo.currentText(), update=False)
        self.ui.selected_region.sigRegionChanged.connect(self.update_measurement)
        self.ui.result_sb.valueChanged.connect(self.ui.result_lcd.display)

    @Slot(dict)
    def update_fft_filter(self, d):
        self.frequency = d['frequency']
        self.phase = d['phase']
        self.update_measurement()

    def Quit_fun(self):
        """
            close the current instance of daq_measurement.

        """
        # insert anything that needs to be closed before leaving
        self.parent.close()

    def update_status(self, txt, wait_time=0):
        """
            Update the statut bar showing the given text message with a delay of wait_time ms (0s by default).

            =============== ========= ===========================
            **Parameters**

            *txt*             string   the text message to show

            *wait_time*       int      the delay time of waiting
            =============== ========= ===========================

        """
        self.ui.statusbar.showMessage(txt, wait_time)

    @Slot(str)
    def update_measurement_subtype(self, mtype, update=True):
        """
            | Update the ui-measure_subtype_combo from subitems and formula attributes, if specified by update parameter.
            | Linked with the update_measurement method

            ================ ========== =====================================================================================
            **Parameters**    **Type**   **Description**

            mtype             string      the Measurement_type index of the Measurement_type array (imported from daq_utils)

            update            boolean     the update boolean link with the update_measurement method
            ================ ========== =====================================================================================

            See Also
            --------
            update_measurement_subtype, update_measurement, update_status

        """
        self.measurement_type = Measurement_type[mtype]
        [variables, self.formula, self.subitems] = Measurement_type.update_measurement_subtype(mtype)

        try:
            self.ui.measure_subtype_combo.clear()
            self.ui.measure_subtype_combo.addItems(self.subitems)
            self.ui.formula_edit.setPlainText(self.formula)

            if update:
                self.update_measurement()
        except Exception as e:
            self.update_status(str(e), wait_time=self.wait_time)

    def update_measurement(self):
        """
            Update :
             * the measurement results from the update_measurements method
             * the statut bar on cascade (if needed)
             * the User Interface function curve state and data (if needed).

            Emit the measurement_signal corresponding.

            See Also
            --------
            update_measurement, update_status

        """
        try:
            xlimits = self.ui.selected_region.getRegion()
            mtype = self.ui.measurement_type_combo.currentText()
            msubtype = self.ui.measure_subtype_combo.currentText()
            if mtype == 'Sinus':

                # boundaries = utils.find_index(self.xdata, [xlimits[0], xlimits[1]])
                # sub_xaxis = self.xdata[boundaries[0][0]:boundaries[1][0]]
                # sub_data = self.ydata[boundaries[0][0]:boundaries[1][0]]
                # self.fourierfilt.parent.setVisible(True)
                self.fourierfilt.show_data(dict(data=self.ydata, xaxis=self.xdata))
            else:
                # self.fourierfilt.parent.setVisible(False)
                pass

            measurement_results = self.do_measurement(xlimits[0], xlimits[1], self.xdata, self.ydata, mtype, msubtype)
            if measurement_results['status'] is not None:
                self.update_status(measurement_results['status'], wait_time=self.wait_time)
                return
            self.ui.result_sb.setValue(measurement_results['value'])
            self.measurement_signal.emit([measurement_results['value']])
            if measurement_results['datafit'] is not None:
                self.ui.fit_curve.setVisible(True)
                self.ui.fit_curve.setData(measurement_results['xaxis'], measurement_results['datafit'])
            else:
                self.ui.fit_curve.setVisible(False)
        except Exception as e:
            self.update_status(str(e), wait_time=self.wait_time)

    def eval_func(self, x, *args):
        dic = dict(zip(self.subitems, args))
        dic.update(dict(np=np, x=x))
        return eval(self.formula, dic)

    def do_measurement(self, xmin, xmax, xaxis, data1D, mtype, msubtype):
        try:
            boundaries = utils.find_index(xaxis, [xmin, xmax])
            sub_xaxis = xaxis[boundaries[0][0]:boundaries[1][0]]
            sub_data = data1D[boundaries[0][0]:boundaries[1][0]]
            mtypes = Measurement_type.names()
            if msubtype in self.subitems:
                msub_ind = self.subitems.index(msubtype)

            measurement_results = dict(status=None, value=0, xaxis=np.array([]), datafit=np.array([]))

            if mtype == 'Cursor_Integration':  # "Cursor Intensity Integration":
                if msubtype == "sum":
                    result_measurement = np.sum(sub_data)
                elif msubtype == "mean":
                    result_measurement = np.mean(sub_data)
                elif msubtype == "std":
                    result_measurement = np.std(sub_data)
                else:
                    result_measurement = 0

            elif mtype == 'Max':  # "Max":
                result_measurement = np.max(sub_data)

            elif mtype == 'Min':  # "Min":
                result_measurement = np.min(sub_data)

            elif mtype == 'Gaussian_Fit':  # "Gaussian Fit":
                measurement_results['xaxis'] = sub_xaxis
                offset = np.min(sub_data)
                amp = np.max(sub_data) - np.min(sub_data)
                m = utils.my_moment(sub_xaxis, sub_data)
                p0 = [amp, m[1], m[0], offset]
                popt, pcov = curve_fit(self.eval_func, sub_xaxis, sub_data, p0=p0)
                measurement_results['datafit'] = self.eval_func(sub_xaxis, *popt)
                result_measurement = popt[msub_ind]

            elif mtype == 'Lorentzian_Fit':  # "Lorentzian Fit":
                measurement_results['xaxis'] = sub_xaxis
                offset = np.min(sub_data)
                amp = np.max(sub_data) - np.min(sub_data)
                m = utils.my_moment(sub_xaxis, sub_data)
                p0 = [amp, m[1], m[0], offset]
                popt, pcov = curve_fit(self.eval_func, sub_xaxis, sub_data, p0=p0)
                measurement_results['datafit'] = self.eval_func(sub_xaxis, *popt)
                if msub_ind == 4:  # amplitude
                    result_measurement = popt[0] * 2 / (np.pi * popt[1])  # 2*alpha/(pi*gamma)
                else:
                    result_measurement = popt[msub_ind]

            elif mtype == 'Exponential_Decay_Fit':  # "Exponential Decay Fit":
                ind_x0 = utils.find_index(sub_data, np.max(sub_data))[0][0]
                x0 = sub_xaxis[ind_x0]
                sub_xaxis = sub_xaxis[ind_x0:]
                sub_data = sub_data[ind_x0:]
                offset = min([sub_data[0], sub_data[-1]])
                measurement_results['xaxis'] = sub_xaxis
                N0 = np.max(sub_data) - offset
                t37 = sub_xaxis[utils.find_index(sub_data - offset, 0.37 * N0)[0][0]] - x0
                # polynome = np.polyfit(sub_xaxis, -np.log((sub_data - 0.99 * offset) / N0), 1)
                p0 = [N0, t37, x0, offset]
                popt, pcov = curve_fit(self.eval_func, sub_xaxis, sub_data, p0=p0)
                measurement_results['datafit'] = self.eval_func(sub_xaxis, *popt)
                result_measurement = popt[msub_ind]

            elif mtype == 'Sinus':  #
                offset = np.mean(sub_data)
                A = (np.max(sub_data) - np.min(sub_data)) / 2
                phi = self.fourierfilt.phase
                dx = 1 / self.fourierfilt.frequency
                measurement_results['xaxis'] = sub_xaxis
                p0 = [A, dx, phi, offset]
                popt, pcov = curve_fit(self.eval_func, sub_xaxis, sub_data, p0=p0)
                measurement_results['datafit'] = self.eval_func(sub_xaxis, *popt)
                result_measurement = popt[msub_ind]

            # elif mtype=="Custom Formula":
            #    #offset=np.min(sub_data)
            #    #amp=np.max(sub_data)-np.min(sub_data)
            #    #m=utils.my_moment(sub_xaxis,sub_data)
            #    #p0=[amp,m[1],m[0],offset]
            #    popt, pcov = curve_fit(self.custom_func, sub_xaxis, sub_data,p0=[140,750,50,15])
            #    self.curve_fitting_sig.emit([sub_xaxis,self.gaussian_func(sub_xaxis,*popt)])
            #    result_measurement=popt[msub_ind]
            else:
                result_measurement = 0

            measurement_results['value'] = result_measurement

            return measurement_results
        except Exception as e:
            result_measurement = 0
            measurement_results['status'] = str(e)
            return measurement_results

    def update_data(self, xdata=None, ydata=None):
        """
            | Update xdata attribute with the numpy linspcae regular distribution (if param is none) and update the User Interface curve data.
            | Call the update_measurement method synchronously to keep same values.

            =============== ============   ======================
            **Parameters**   **Type**       **Description**

            *xdata*          float list    the x axis data
            *ydata*          float list    the y axis data
            =============== ============   ======================

            See Also
            --------
            update_measurement

        """
        if xdata is None:
            self.xdata = np.linspace(0, len(ydata) - 1, len(ydata))
        else:
            self.xdata = xdata
        self.ydata = ydata
        self.fourierfilt.show_data(dict(data=self.ydata, xaxis=self.xdata))
        self.update_measurement()


if __name__ == '__main__':
    app = QtWidgets.QApplication(sys.argv)
    Form = QtWidgets.QWidget()
    from pymodaq.utils.daq_utils import gauss1D

    prog = DAQ_Measurement(Form)
    xdata = np.linspace(0, 400, 401)
    x0 = 50
    dx = 20
    tau = 27
    tau2 = 100
    ydata_gauss = 10 * gauss1D(xdata, x0, dx) + np.random.rand(len(xdata))
    ydata_expodec = np.zeros((len(xdata)))
    ydata_expodec[:50] = 10 * gauss1D(xdata[:50], x0, dx, 2)
    ydata_expodec[50:] = 10 * np.exp(-(xdata[50:] - x0) / tau)  # +10*np.exp(-(xdata[50:]-x0)/tau2)
    ydata_expodec += 2 * np.random.rand(len(xdata))
    ydata_sin = 10 + 2 * np.sin(2 * np.pi * xdata / 21 - np.deg2rad(55)) + 2 * np.random.rand(len(xdata))
    prog.update_data(xdata, ydata_sin)
    Form.show()

    sys.exit(app.exec_())



from typing import List, Union, TYPE_CHECKING

from collections import OrderedDict
from qtpy.QtCore import QObject, Signal, Slot, QThread
from qtpy import QtWidgets
import time

from pymodaq.utils.logger import set_logger, get_module_name, get_module_name
from pymodaq.utils import daq_utils as utils
from pymodaq.utils.config import Config
from pymodaq.utils.data import DataToExport, DataFromPlugins, DataActuator
from pyqtgraph.parametertree import Parameter, ParameterTree
from pymodaq.utils.managers.parameter_manager import ParameterManager


if TYPE_CHECKING:
    from pymodaq.control_modules.daq_viewer import DAQ_Viewer
    from pymodaq.control_modules.daq_move import DAQ_Move

logger = set_logger(get_module_name(__file__))
config = Config()


class ModulesManager(QObject, ParameterManager):
    """Class to manage DAQ_Viewers and DAQ_Moves with UI to select some

    Easier to connect control modules signals to slots, test, ...

    Parameters
    ----------
    detectors: list of DAQ_Viewer
    actuators: list of DAQ_Move
    selected_detectors: list of DAQ_Viewer
        sublist of detectors
    selected_actuators: list of DAQ_Move
        sublist of actuators
    """
    detectors_changed = Signal(list)
    actuators_changed = Signal(list)
    det_done_signal = Signal(DataToExport)  # dte here contains DataWithAxes
    move_done_signal = Signal(DataToExport)  # dte here contains DataActuators
    timeout_signal = Signal(bool)

    params = [
        {'title': 'Actuators/Detectors Selection', 'name': 'modules', 'type': 'group', 'children': [
            {'title': 'detectors', 'name': 'detectors', 'type': 'itemselect', 'checkbox': True},
            {'title': 'Actuators', 'name': 'actuators', 'type': 'itemselect', 'checkbox': True},
        ]},

        {'title': "Moves done?", 'name': 'move_done', 'type': 'led', 'value': False},
        {'title': "Detections done?", 'name': 'det_done', 'type': 'led', 'value': False},

        {'title': 'Data dimensions', 'name': 'data_dimensions', 'type': 'group', 'children': [
            {'title': "Probe detector's data", 'name': 'probe_data', 'type': 'action'},
            {'title': 'Data0D list:', 'name': 'det_data_list0D', 'type': 'itemselect'},
            {'title': 'Data1D list:', 'name': 'det_data_list1D', 'type': 'itemselect'},
            {'title': 'Data2D list:', 'name': 'det_data_list2D', 'type': 'itemselect'},
            {'title': 'DataND list:', 'name': 'det_data_listND', 'type': 'itemselect'},
        ]},
        {'title': 'Actuators positions', 'name': 'actuators_positions', 'type': 'group', 'children': [
            {'title': "Test actuators", 'name': 'test_actuator', 'type': 'action'},
            {'title': 'Positions:', 'name': 'positions_list', 'type': 'itemselect'},
        ]},
    ]

    def __init__(self, detectors=[], actuators=[], selected_detectors=[], selected_actuators=[], **kwargs):
        QObject.__init__(self)
        ParameterManager.__init__(self)

        for mod in selected_actuators:
            assert mod in actuators
        for mod in selected_detectors:
            assert mod in detectors

        self.actuator_timeout = config('actuator', 'timeout')
        self.detector_timeout = config('viewer', 'timeout')

        self.det_done_datas: DataToExport = None
        self.det_done_flag = False
        self.move_done_positions: DataToExport = None
        self.move_done_flag = False

        self.settings.child('data_dimensions', 'probe_data').sigActivated.connect(self.get_det_data_list)
        self.settings.child('actuators_positions', 'test_actuator').sigActivated.connect(self.test_move_actuators)

        self._detectors = []
        self._actuators = []

        self.grab_done_signals = []
        self.det_commands_signal = []

        self.actuators_connected = False
        self.detectors_connected = False

        self.set_actuators(actuators, selected_actuators)
        self.set_detectors(detectors, selected_detectors)

    def show_only_control_modules(self, show: True):
        self.settings.child('move_done').show(not show)
        self.settings.child('det_done').show(not show)
        self.settings.child('data_dimensions').show(not show)
        self.settings.child('actuators_positions').show(not show)

    @classmethod
    def get_names(cls, modules):
        """Get the titles of a list of Control Modules

        Parameters
        ----------
        modules: list of DAQ_Move and/or DAQ_Viewer
        """
        if not hasattr(modules, '__iter__'):
            modules = [modules]
        return [mod.title for mod in modules]

    def get_mods_from_names(self, names, mod='det'):
        """Getter of a list of given modules from their name (title)

        Parameters
        ----------
        names: list of str
        mod: str
            either 'det' for DAQ_Viewer modules or 'act' for DAQ_Move modules
        """
        mods = []
        for name in names:
            d = self.get_mod_from_name(name, mod)
            if d is not None:
                mods.append(d)
        return mods

    def get_mod_from_name(self, name, mod='det') -> Union['DAQ_Move', 'DAQ_Viewer']:
        """Getter of a given module from its name (title)

        Parameters
        ----------
        name: str
        mod: str
            either 'det' for DAQ_Viewer modules or 'act' for DAQ_Move modules
        """
        if mod == 'det':
            modules = self._detectors
        else:
            modules = self._actuators

        if name in self.get_names(modules):
            return modules[self.get_names(modules).index(name)]
        else:
            logger.warning(f'No detector with this name: {name}')
            return None

    def set_actuators(self, actuators, selected_actuators):
        """Populates actuators and the subset to be selected in the UI"""
        self._actuators = actuators
        self.settings.child('modules', 'actuators').setValue(dict(all_items=self.get_names(actuators),
                                                                  selected=self.get_names(selected_actuators)))

    def set_detectors(self, detectors, selected_detectors):
        """Populates detectors and the subset to be selected in the UI"""
        self._detectors = detectors
        self.settings.child('modules', 'detectors').setValue(dict(all_items=self.get_names(detectors),
                                                                  selected=self.get_names(selected_detectors)))

    @property
    def detectors(self) -> List['DAQ_Viewer']:
        """Get the list of selected detectors"""
        return self.get_mods_from_names(self.selected_detectors_name)

    @property
    def detectors_all(self):
        """Get the list of all detectors"""
        return self._detectors

    @property
    def actuators(self) -> List['DAQ_Move']:
        """Get the list of selected actuators"""
        return self.get_mods_from_names(self.selected_actuators_name, mod='act')

    @property
    def actuators_all(self):
        """Get the list of all actuators"""
        return self._actuators

    @property
    def modules(self):
        """Get the list of all detectors and actuators"""
        return self.detectors + self.actuators

    @property
    def modules_all(self):
        """Get the list of all detectors and actuators"""
        return self.detectors_all + self.actuators_all

    @property
    def Ndetectors(self):
        """Get the number of selected detectors"""
        return len(self.detectors)

    @property
    def Nactuators(self):
        """Get the number of selected actuators"""
        return len(self.actuators)

    @property
    def detectors_name(self):
        """Get all the names of the detectors"""
        return self.settings.child('modules', 'detectors').value()['all_items']

    @property
    def selected_detectors_name(self):
        """Get/Set the names of the selected detectors"""
        return self.settings.child('modules', 'detectors').value()['selected']

    @selected_detectors_name.setter
    def selected_detectors_name(self, detectors):
        if set(detectors).issubset(self.detectors_name):
            self.settings.child('modules', 'detectors').setValue(dict(all_items=self.detectors_name,
                                                                      selected=detectors))

    @property
    def actuators_name(self):
        """Get all the names of the actuators"""
        return self.settings.child('modules', 'actuators').value()['all_items']

    @property
    def selected_actuators_name(self) -> List[str]:
        """Get/Set the names of the selected actuators"""
        return self.settings.child('modules', 'actuators').value()['selected']

    @selected_actuators_name.setter
    def selected_actuators_name(self, actuators):
        if set(actuators).issubset(self.actuators_name):
            self.settings.child('modules', 'actuators').setValue(dict(all_items=self.actuators_name,
                                                                      selected=actuators))

    def value_changed(self, param):
        if param.name() == 'detectors':
            self.detectors_changed.emit(param.value()['selected'])

        elif param.name() == 'actuators':
            self.actuators_changed.emit(param.value()['selected'])

    def get_det_data_list(self):
        """Do a snap of selected detectors, to get the list of all the data and processed data"""

        self.connect_detectors()
        datas: DataToExport = self.grab_datas()

        data_list0D = datas.get_full_names('data0D')
        data_list1D = datas.get_full_names('data1D')
        data_list2D = datas.get_full_names('data2D')
        data_listND = datas.get_full_names('dataND')

        self.settings.child('data_dimensions', 'det_data_list0D').setValue(
            dict(all_items=data_list0D, selected=[]))
        self.settings.child('data_dimensions', 'det_data_list1D').setValue(
            dict(all_items=data_list1D, selected=[]))
        self.settings.child('data_dimensions', 'det_data_list2D').setValue(
            dict(all_items=data_list2D, selected=[]))
        self.settings.child('data_dimensions', 'det_data_listND').setValue(
            dict(all_items=data_listND, selected=[]))

        self.connect_detectors(False)

    def get_selected_probed_data(self, dim='0D'):
        """Get the name of selected data names of a given dimensionality

        Parameters
        ----------
        dim: str
            either '0D', '1D', '2D' or 'ND'
        """
        return self.settings.child('data_dimensions', f'det_data_list{dim.upper()}').value()['selected']

    def grab_datas(self, **kwargs):
        """Do a single grab of connected and selected detectors"""
        self.det_done_datas = DataToExport(name=__class__.__name__, control_module='DAQ_Viewer')
        self._received_data = 0
        self.det_done_flag = False
        self.settings.child('det_done').setValue(self.det_done_flag)
        tzero = time.perf_counter()

        for mod in self.detectors:
            kwargs.update(dict(Naverage=mod.Naverage))
            mod.command_hardware.emit(utils.ThreadCommand("single", kwargs))

        while not self.det_done_flag:
            # wait for grab done signals to end

            QtWidgets.QApplication.processEvents()
            if time.perf_counter() - tzero > self.detector_timeout:
                self.timeout_signal.emit(True)
                logger.error('Timeout Fired during waiting for data to be acquired')
                break

        self.det_done_signal.emit(self.det_done_datas)
        return self.det_done_datas

    def connect_actuators(self, connect=True, slot=None, signal='move_done'):
        """Connect the selected actuators signal to a given or default slot

        Parameters
        ----------
        connect: bool
        slot: builtin_function_or_method
            method or function the chosen signal will be connected to
            if None, then the default move_done slot is used
        signal: str
            What kind of signal is to be used:

            * 'move_done' will connect the `move_done_signal` to the slot
            * 'current_value' will connect the 'current_value_signal' to the slot

        See Also
        --------
        :meth:`move_done`
        """
        if slot is None:
            slot = self.move_done
        if connect:
            for sig in [mod.move_done_signal if signal == 'move_done' else mod.current_value_signal
                        for mod in self.actuators]:
                sig.connect(slot)

        else:
            try:
                for sig in [mod.move_done_signal if signal == 'move_done' else mod.current_value_signal
                            for mod in self.actuators]:
                    sig.disconnect(slot)
            except Exception as e:
                logger.error(str(e))

        self.actuators_connected = connect

    def connect_detectors(self, connect=True, slot=None):
        """
        Connect selected DAQ_Viewers's grab_done_signal to the given slot

        Parameters
        ----------
        connect: bool
            if True, connect to the given slot (or default slot)
            if False, disconnect all detectors (not only the currently selected ones.
            This is made because when selected detectors changed if you only disconnect those one,
            the previously connected ones will stay connected)
        slot: method
            A method that should be connected, if None self.det_done is connected by default
        """

        if slot is None:
            slot = self.det_done

        if connect:
            for sig in [mod.grab_done_signal for mod in self.detectors]:
                sig.connect(slot)
        else:

            for sig in [mod.grab_done_signal for mod in self.detectors_all]:
                try:
                    sig.disconnect(slot)
                except TypeError as e:
                    # means the slot was not previously connected
                    logger.info(str(e))

        self.detectors_connected = connect

    def test_move_actuators(self):
        """Do a move of selected actuator"""
        dte_act = DataToExport('Actuators', control_module='DAQ_MOVE')
        for act in self.get_names(self.actuators):
            pos, done = QtWidgets.QInputDialog.getDouble(None, f'Enter a target position for actuator {act}',
                                                         'Position:')
            if not done:
                pos = 0.
            dte_act.append(DataActuator(act, data=pos))

        self.connect_actuators()

        self.move_actuators(dte_act)

        self.settings.child('actuators_positions',
                            'positions_list').setValue(dict(all_items=[f'{dact.name}: {dact.value()}' for dact
                                                                       in dte_act],
                                                            selected=[]))

        self.connect_actuators(False)

    def move_actuators(self, dte_act: DataToExport, mode='abs', polling=True) -> DataToExport:
        """will apply positions to each currently selected actuators. By Default the mode is absolute but can be

        Parameters
        ----------
        dte_act: DataToExport
            the DataToExport of position to apply. Its length must be equal to the number of selected actuators
        mode: str
            either 'abs' for absolute positionning or 'rel' for relative
        polling: bool
            if True will wait for the selected actuators to reach their target positions (they have to be
            connected to a method checking for the position and letting the programm know the move is done (default
            connection is this object `move_done` method)

        Returns
        -------
        DataToExport with the selected actuators's name as key and current actuators's value as value
        """
        self.move_done_positions = DataToExport(name=__class__.__name__, control_module='DAQ_Move')
        self.move_done_flag = False
        self.settings.child('move_done').setValue(self.move_done_flag)

        if mode == 'abs':
            command = 'move_abs'
        elif mode == 'rel':
            command = 'move_rel'
        else:
            logger.error(f'Invalid positioning mode: {mode}')
            return self.move_done_positions

        if len(dte_act) == self.Nactuators:
            for dact in dte_act:
                act = self.get_mod_from_name(dact.name, 'act')
                if act is not None:
                    act.command_hardware.emit(
                        utils.ThreadCommand(command=command, attribute=[dact, polling]))
            # else:
            #     for ind, act in enumerate(self.actuators):
            #         #getattr(act, command)(positions[ind])
            #         act.command_hardware.emit(utils.ThreadCommand(command=command, attribute=[positions[ind], polling]))

        else:
            logger.error('Invalid number of positions compared to selected actuators')
            return self.move_done_positions

        tzero = time.perf_counter()
        if polling:
            while not self.move_done_flag:  # polling move done

                QtWidgets.QApplication.processEvents()
                if time.perf_counter() - tzero > self.actuator_timeout / 1000:  # timeout in seconds
                    self.timeout_signal.emit(True)
                    logger.error('Timeout Fired during waiting for actuators to be moved')
                    break
                QThread.msleep(20)

        self.move_done_signal.emit(self.move_done_positions)
        return self.move_done_positions

    def reset_signals(self):
        self.move_done_flag = True
        self.det_done_flag = True

    def order_positions(self, positions: DataToExport):
        """ Reorder the content of the DataToExport given the order of the selected actuators"""
        actuators = self.selected_actuators_name
        pos = DataToExport('actuators')
        for act in actuators:
            pos.append(positions.get_data_from_name(act))
        return pos

    @Slot(DataActuator)
    def move_done(self, data_act: DataActuator):
        try:
            if data_act.name not in self.move_done_positions.get_names():
                self.move_done_positions.append(data_act)

            if len(self.move_done_positions) == len(self.actuators):
                self.move_done_flag = True
                self.settings.child('move_done').setValue(self.move_done_flag)
        except Exception as e:
            logger.exception(str(e))

    def det_done(self, data: DataToExport):
        if self.det_done_datas is not None:  # means that somehow data are not initialized so no further processing
            self._received_data += 1
            if len(data) != 0:
                self.det_done_datas.append(data)

            if self._received_data == len(self.detectors):
                self.det_done_flag = True
                self.settings.child('det_done').setValue(self.det_done_flag)

        # if data.name not in list(self.det_done_datas.keys()):
        #     self.det_done_datas[data['name']] = data
        # if len(self.det_done_datas.items()) == len(self.detectors):
        #     self.det_done_flag = True


if __name__ == '__main__':
    import sys

    app = QtWidgets.QApplication(sys.argv)
    from qtpy.QtCore import QThread
    from pymodaq.utils.gui_utils import DockArea
    from pyqtgraph.dockarea import Dock
    from pymodaq.control_modules.daq_viewer import DAQ_Viewer
    from pymodaq.control_modules.daq_move import DAQ_Move

    win = QtWidgets.QMainWindow()
    area = DockArea()
    win.setCentralWidget(area)
    win.resize(1000, 500)
    win.setWindowTitle('pymodaq main')

    prog = DAQ_Viewer(area, title="Testing2D", daq_type='DAQ2D')
    prog2 = DAQ_Viewer(area, title="Testing1D", daq_type='DAQ1D')
    prog3 = DAQ_Viewer(area, title="Testing0D", daq_type='DAQ0D')

    act1_widget = QtWidgets.QWidget()
    act2_widget = QtWidgets.QWidget()
    act1 = DAQ_Move(act1_widget, title='X_axis')
    act2 = DAQ_Move(act2_widget, title='Y_axis')
    act1.actuator = 'Mock'
    act2.actuator = 'Mock'

    QThread.msleep(1000)
    prog.init_hardware_ui()
    prog2.init_hardware_ui()
    prog3.init_hardware_ui()

    dock1 = Dock('actuator 1')
    dock1.addWidget(act1_widget)
    area.addDock(dock1)

    dock2 = Dock('actuator 2')
    dock2.addWidget(act2_widget)
    area.addDock(dock2)

    act1.init_hardware_ui()
    act2.init_hardware_ui()

    QtWidgets.QApplication.processEvents()
    win.show()

    manager = ModulesManager(actuators=[act1, act2], detectors=[prog, prog2, prog3], selected_detectors=[prog2])
    manager.settings_tree.show()

    sys.exit(app.exec_())



import logging
import datetime
from typing import List

from pymodaq.utils.logger import set_logger, get_module_name, get_module_name
from pymodaq.utils.config import Config
from pymodaq.utils.data import DataToExport
from qtpy import QtCore
from contextlib import contextmanager

from pymodaq.utils.db.db_logger.db_logger_models import Base, Data0D, Data1D, Data2D, LogInfo,\
    Configuration, ControlModule
from pymodaq.utils import daq_utils as utils
from pymodaq.utils.gui_utils.utils import dashboard_submodules_params
from pymodaq.utils.messenger import messagebox, deprecation_msg
from pymodaq.utils.abstract.logger import AbstractLogger
from pymodaq.utils.parameter import ParameterTree, Parameter
from pymodaq.utils.managers.parameter_manager import ParameterManager

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy_utils import database_exists, create_database

logger = set_logger(get_module_name(__file__))
config = Config()


class DBLogHandler(logging.StreamHandler):
    def __init__(self, dblogger):
        super().__init__()
        self.dblogger = dblogger
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        self.setFormatter(formatter)

    def emit(self, record):
        msg = self.format(record)
        self.dblogger.add_log(msg)


class DbLogger:
    user = config('network', 'logging', 'user', 'username')
    user_pwd = config('network', 'logging', 'user', 'pwd')

    def __init__(self, database_name, ip_address=config('network', 'logging', 'sql', 'ip'),
                 port=config('network', 'logging', 'sql', 'port'), save2D=False):
        """

        Parameters
        ----------
        models_base
        ip_address
        port
        database_name
        """

        self.ip_address = ip_address
        self.port = port
        self.database_name = database_name

        self.engine = None
        self.Session = None
        self._save2D = save2D

    @property
    def save2D(self):
        return self._save2D

    @save2D.setter
    def save2D(self, value):
        self._save2D = value

    @contextmanager
    def session_scope(self):
        """Provide a transactional scope around a series of operations."""
        session = self.Session()
        try:
            yield session
            session.commit()
        except Exception as e:
            logger.error(str(e))
            session.rollback()
        finally:
            session.close()

    def connect_db(self):
        url = f"postgresql://{self.user}:{self.user_pwd}@{self.ip_address}:"\
              f"{self.port}/{self.database_name}"
        logger.debug(f'Connecting database using: {url}')
        try:
            self.engine = create_engine(f"postgresql://{self.user}:{self.user_pwd}@{self.ip_address}:"
                                        f"{self.port}/{self.database_name}")
        except ModuleNotFoundError as e:
            messagebox('warning', 'ModuleError',
                       f'The postgresql backend *psycopg2* has not been installed.\n'
                       f'Could not connect to your database')
            logger.warning(str(e))
            return False

        try:
            if not database_exists(self.engine.url):
                create_database(self.engine.url)
            assert database_exists(self.engine.url)
        except Exception as e:
            logger.critical(f'Impossible to connect to the DB: {str(e)}')
            return False

        self.create_table()
        self.Session = sessionmaker(bind=self.engine)
        logger.debug(f'Database Connected')
        return True

    def close(self):
        if self.engine is not None:
            self.engine.dispose()

    def create_table(self):
        # create tables if not existing
        if self.engine is not None:
            Base.metadata.create_all(self.engine)

    def get_detectors(self, session) -> List[str]:
        """Returns the list of detectors name

        Parameters
        ----------
        session: (Session) SQLAlchemy session instance for db transactions

        Returns
        -------
        the list of all created detectors
        """
        return [res[0] for res in session.query(ControlModule.name)]

    def get_actuators(self, session) -> List[str]:
        """Returns the list of actuators name

        Parameters
        ----------
        session: (Session) SQLAlchemy session instance for db transactions

        Returns
        -------
        the list of all created actuators
        """
        return [res[0] for res in session.query(ControlModule.name)]

    def add_detectors(self, detectors):
        """
        add detectors in the detectors table
        Parameters
        ----------
        detectors: (list) list of dict with keys: name and settings_xml
        """
        self.add_control_modules(detectors, 'DAQ_Viewer')

    def add_actuators(self, actuators):
        """
        add actuators in the actuators table
        Parameters
        ----------
        actuators: list
            list of dict with keys: name and settings_xml
        """
        self.add_control_modules(actuators, 'DAQ_Move')

    def add_control_modules(self, modules, module_type='DAQ_Viewer'):
        if not isinstance(modules, list):
            modules = [modules]
        with self.session_scope() as session:
            existing_modules = [d.name for d in session.query(ControlModule)]
            for mod in modules:
                if mod['name'] not in existing_modules:
                    session.add(ControlModule(name=mod['name'], module_type=module_type,
                                              settings_xml=mod['xml_settings']))

    def add_config(self, config_settings):
        with self.session_scope() as session:
            session.add(Configuration(timestamp=datetime.datetime.now().timestamp(), settings_xml=config_settings))

    def add_log(self, log):
        with self.session_scope() as session:
            session.add(LogInfo(log))

    def add_data(self, data: DataToExport):
        with self.session_scope() as session:
            time_stamp = data.timestamp
            module_name = data.name

            if session.query(ControlModule).filter_by(name=module_name).count() == 0:
                self.add_control_modules(session, dict(name=module_name), data.control_module)

            module_id = session.query(ControlModule).filter_by(name=module_name).one().id  # detector/actuator names should/are unique

            for dwa in data.get_data_from_dim('Data0D'):
                for ind, data_array in enumerate(dwa):
                    session.add(Data0D(timestamp=dwa.timestamp, control_module_id=module_id, channel=dwa.labels[ind],
                                       value=float(data_array[0])))

            for dwa in data.get_data_from_dim('Data1D'):
                for ind, data_array in enumerate(dwa):
                    session.add(Data1D(timestamp=dwa.timestamp, control_module_id=module_id,
                                       channel=dwa.labels[ind],
                                       value=data_array.tolist()))
            if self.save2D:
                for dwa in data.get_data_from_dim('Data2D'):
                    for ind, data_array in enumerate(dwa):
                        session.add(Data2D(timestamp=dwa.timestamp, control_module_id=module_id,
                                           channel=dwa.labels[ind],
                                           value=data_array.tolist()))

            # not yet dataND as db should not know where to save these datas


class DbLoggerGUI(DbLogger, ParameterManager):
    params = [
        {'title': 'Database:', 'name': 'database_type', 'type': 'list', 'value': 'PostgreSQL',
            'limits': ['PostgreSQL', ]},
        {'title': 'Server IP:', 'name': 'server_ip', 'type': 'str',
            'value': config('network', 'logging', 'sql', 'ip'),
         'tip':'Either localhost if the database server is on the same computer or the IP address of the server'},
        {'title': 'Server port:', 'name': 'server_port', 'type': 'int',
            'value': config('network', 'logging', 'sql', 'port')},
        {'title': 'Connect:', 'name': 'connect_db', 'type': 'bool_push', 'value': False},
        {'title': 'Connected:', 'name': 'connected_db', 'type': 'led', 'value': False},
    ] + dashboard_submodules_params

    def __init__(self, database_name):
        DbLogger.__init__(self, database_name, ip_address=config('network', 'logging', 'sql', 'ip'),
                          port=config('network', 'logging', 'sql', 'port'), save2D=False)
        ParameterManager.__init__(self)

        self.settings.child('do_save').hide()

    def value_changed(self, param):
        if param.name() == 'server_ip':
            self.ip_address = param.value()

        elif param.name() == 'server_port':
            self.port = param.value()

        elif param.name() == 'connect_db':
            if param.value():
                status = self.connect_db()
                self.settings.child('connected_db').setValue(status)
            else:
                self.close()
                self.settings.child('connected_db').setValue(False)

        elif param.name() == 'save_2D':
            self.save2D = param.value()


class DataBaseLogger(AbstractLogger):
    def __init__(self, database_name):
        self.dblogger = DbLoggerGUI(database_name)

    @property
    def settings_tree(self):
        return self.dblogger.settings_tree

    @property
    def settings(self):
        return self.dblogger.settings

    def init_logger(self, settings):
        if not self.settings.child('connected_db').value():
            status = self.dblogger.connect_db()
            if not status:
                logger.critical('the Database is not and cannot be connnect')
                return False

        self.dblogger.add_config(settings)
        return True

    def get_handler(self):
        return DBLogHandler(self.dblogger)

    def add_detector(self, det_name, settings):
        self.dblogger.add_detectors([dict(name=det_name, xml_settings=settings)])

    def add_actuator(self, act_name, settings):
        self.dblogger.add_actuators([dict(name=act_name, xml_settings=settings)])

    def add_data(self, data):
        self.dblogger.add_data(data)
        self.settings.child('N_saved').setValue(
            self.settings.child('N_saved').value() + 1)

    def stop_logger(self):
        pass

    def close(self):
        pass


if __name__ == '__main__':
    db = DbLogger('preset_default')
    db.connect_db()
    pass


import os
from pathlib import Path
from qtpy import QtWidgets, QtCore, QtGui
from pyqtgraph.parametertree.Parameter import ParameterItem
from pyqtgraph.parametertree.parameterTypes.basetypes import WidgetParameterItem
from pyqtgraph.parametertree import Parameter


class FileDirWidget(QtWidgets.QWidget):
    """
        ================ =========================
        **Attributes**    **Type**
        *value_changed*   instance of pyqt Signal
        *path*            string
        ================ =========================

        See Also
        --------
        browse_path
    """
    value_changed = QtCore.Signal(str)

    def __init__(self, init_path='D:/Data', file_type=False):

        super().__init__()
        self.filetype = file_type
        self.path = init_path
        self.initUI()

        self.base_path_browse_pb.clicked.connect(self.browse_path)

    def browse_path(self):
        """
            Browse the path attribute if exist.

            See Also
            --------
            set_path
        """
        if self.filetype is True:
            folder_name = QtWidgets.QFileDialog.getOpenFileName(None, 'Choose File', os.path.split(self.path)[0])[0]
        elif self.filetype is False:
            folder_name = QtWidgets.QFileDialog.getExistingDirectory(None, 'Choose Folder', self.path)

        elif self.filetype == "save":
            folder_name = QtWidgets.QFileDialog.getSaveFileName(None, 'Enter a Filename', os.path.split(self.path)[0])[
                0]

        if not (not(folder_name)):  # execute if the user didn't cancel the file selection
            self.set_path(folder_name)
            self.value_changed.emit(folder_name)

    def set_path(self, path_file):
        """
            Set the base path attribute with the given path_file.

            =============== =========== ===========================
            **Parameters**    **Type**    **Description**
            *path_file*       string      the pathname of the file
            =============== =========== ===========================
        """
        if isinstance(path_file, Path):
            path_file = str(path_file)
        self.base_path_edit.setPlainText(path_file)
        self.path = path_file

    def get_value(self):
        """
            Get the value of the base_path_edit attribute.

            Returns
            -------
            string
                the path name
        """
        return self.base_path_edit.toPlainText()

    def initUI(self):
        """
            Init the User Interface.
        """

        self.hor_layout = QtWidgets.QHBoxLayout()
        self.base_path_edit = QtWidgets.QPlainTextEdit(self.path)
        self.base_path_edit.setMaximumHeight(50)
        self.base_path_browse_pb = QtWidgets.QPushButton()
        self.base_path_browse_pb.setText("")
        icon3 = QtGui.QIcon()
        icon3.addPixmap(QtGui.QPixmap(":/icons/Icon_Library/Browse_Dir_Path.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.base_path_browse_pb.setIcon(icon3)
        self.hor_layout.addWidget(self.base_path_edit)

        verlayout = QtWidgets.QVBoxLayout()
        verlayout.addWidget(self.base_path_browse_pb)
        verlayout.addStretch()
        self.hor_layout.addLayout(verlayout)
        self.hor_layout.setSpacing(0)
        self.setLayout(self.hor_layout)


class FileDirParameterItem(WidgetParameterItem):
    def makeWidget(self):
        """
            Make an initialized file_browser object with parameter options dictionnary ('readonly' key)0

            Returns
            -------
            w : filebrowser
                The initialized file browser.

            See Also
            --------
            file_browser
        """
        self.asSubItem = True
        self.hideWidget = False
        if 'filetype' in self.param.opts:
            self.filetype = self.param.opts['filetype']
        else:
            self.filetype = True

        self.w = FileDirWidget(self.param.value(), file_type=self.filetype)

        # if 'tip' in self.param.opts:
        #     self.w.setToolTip(self.param.opts['tip'])

        self.w.base_path_edit.setReadOnly(self.param.opts['readonly'])
        self.w.value = self.w.get_value
        self.w.setValue = self.w.set_path
        self.w.sigChanged = self.w.value_changed
        return self.w


class FileDirParameter(Parameter):
    """
        Editable string; displayed as large text box in the tree.
        See Also
        --------
        file_browserParameterItem
    """
    itemClass = FileDirParameterItem





from qtpy import QtGui, QtWidgets
from qtpy.QtCore import Qt, QObject, QTimer
from pyqtgraph.dockarea import Dock
from pymodaq.utils.gui_utils import DockArea
import time
from datetime import timedelta
import sys
from pymodaq.resources.QtDesigner_Ressources import QtDesigner_ressources_rc


class PushButtonShortcut(QtWidgets.QPushButton):

    def __init__(self, *args, shortcut=None, shortcut_widget=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.shortcut = None
        if shortcut_widget is None:
            shortcut_widget = self
        if shortcut is not None:
            self.shortcut = QtWidgets.QShortcut(QtGui.QKeySequence(shortcut), shortcut_widget)
            self.shortcut.activated.connect(self.click_button)

    def click_button(self):
        self.click()


class ChronoTimer(QObject):

    def __init__(self, parent, duration=None):
        """

        :param parent:
        :param Nteams:
        :param duration: (dict) containing optional keys : days, minutes, seconds, hours, weeks, milliseconds, microseconds
        """
        super().__init__()
        self.area = parent
        if duration is None:
            self.type = 'chrono'
            self.duration = timedelta()
        else:
            self.type = 'timer'
            self.duration = timedelta(**duration)  # seconds
        self.displayed_time = 0  # in seconds
        self.started = False
        self.timer = QTimer()
        self.timer.setInterval(100)
        self.timer.timeout.connect(self.display_time)

        self.setup_ui()

    def setup_ui(self):

        self.dock_chrono_timer = Dock(self.type.capitalize())
        self.area.addDock(self.dock_chrono_timer)
        self.dock_chrono_timer.float()

        widget_chrono_timer = QtWidgets.QWidget()
        self.dock_chrono_timer.addWidget(widget_chrono_timer)

        self.layout_lcd = QtWidgets.QVBoxLayout()
        widget_chrono_timer.setLayout(self.layout_lcd)

        self.dock_chrono_timer.setAutoFillBackground(True)
        palette = self.dock_chrono_timer.palette()
        palette.setColor(palette.Background, QtGui.QColor(0, 0, 0))
        self.dock_chrono_timer.setPalette(palette)

        self.time_lcd = QtWidgets.QLCDNumber(8)
        self.set_lcd_color(self.time_lcd, 'red')
        self.layout_lcd.addWidget(self.time_lcd)

        hours, minutes, seconds = self.get_times(self.duration)

        self.time_lcd.display(
            '{:02d}:{:02d}:{:02d}'.format(hours, minutes, seconds))

        self.dock_controls = Dock('Chrono/Timer Controls')
        self.area.addDock(self.dock_controls)
        self.dock_controls.setOrientation('vertical', True)
        self.dock_controls.setMaximumHeight(150)

        self.widget_controls = QtWidgets.QWidget()
        self.controls_layout = QtWidgets.QVBoxLayout()
        self.widget_controls.setLayout(self.controls_layout)

        hor_layout = QtWidgets.QHBoxLayout()
        hor_widget = QtWidgets.QWidget()
        hor_widget.setLayout(hor_layout)

        self.controls_layout.addWidget(hor_widget)

        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/icons/Icon_Library/run2.png"), QtGui.QIcon.Normal,
                       QtGui.QIcon.Off)
        self.start_pb = PushButtonShortcut(icon, 'Start',
                                           shortcut='Home', shortcut_widget=self.area)
        self.start_pb.clicked.connect(self.start)
        self.start_pb.setToolTip('home ("début") key shorcut')

        hor_layout.addWidget(self.start_pb)

        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/icons/Icon_Library/pause.png"), QtGui.QIcon.Normal,
                       QtGui.QIcon.Off)
        self.pause_pb = PushButtonShortcut(icon, 'Pause',
                                           shortcut='Ctrl+p', shortcut_widget=self.area)
        self.pause_pb.setCheckable(True)
        self.pause_pb.setToolTip("Ctrl+p key shortcut")
        self.pause_pb.clicked.connect(self.pause)
        hor_layout.addWidget(self.pause_pb)

        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/icons/Icon_Library/Refresh2.png"), QtGui.QIcon.Normal,
                       QtGui.QIcon.Off)
        self.reset_pb = PushButtonShortcut(icon, 'Reset',
                                           shortcut='F5', shortcut_widget=self.area)
        self.reset_pb.setToolTip('F5 key shortcut')
        self.reset_pb.clicked.connect(self.reset)
        hor_layout.addWidget(self.reset_pb)

        self.dock_controls.addWidget(self.widget_controls)

    def get_times(self, duration):
        seconds = int(duration.total_seconds() % 60)
        total_minutes = duration.total_seconds() // 60
        minutes = int(total_minutes % 60)
        hours = int(total_minutes // 60)

        return hours, minutes, seconds

    def get_elapsed_time(self):
        return time.perf_counter() - self.ini_time

    def display_time(self):
        elapsed_time = self.get_elapsed_time()
        if self.type == 'timer':
            display_timedelta = self.duration - timedelta(seconds=elapsed_time)
        else:
            display_timedelta = self.duration + timedelta(seconds=elapsed_time)

        self.displayed_time = display_timedelta.total_seconds()
        if display_timedelta.total_seconds() <= 0:
            self.reset()
            return
        else:
            hours, minutes, seconds = self.get_times(display_timedelta)

            self.time_lcd.display(
                '{:02d}:{:02d}:{:02d}'.format(hours, minutes, seconds))
            QtWidgets.QApplication.processEvents()

    def start(self):
        self.ini_time = time.perf_counter()
        self.timer.start()
        self.started = True
        self.start_pb.setEnabled(False)

    def pause(self):
        if self.pause_pb.isChecked():
            self.started = False
            self.timer.stop()
            self.paused_time = time.perf_counter()
        else:
            elapsed_pause_time = time.perf_counter() - self.paused_time
            self.ini_time += elapsed_pause_time
            self.timer.start()
            self.started = True

    def reset(self):
        if self.pause_pb.isChecked():
            self.pause_pb.setChecked(False)
            QtWidgets.QApplication.processEvents()
        self.timer.stop()
        self.start_pb.setEnabled(True)
        self.started = False
        hours, minutes, seconds = self.get_times(self.duration)
        self.time_lcd.display(
            '{:02d}:{:02d}:{:02d}'.format(hours, minutes, seconds))

    def set_lcd_color(self, lcd, color):
        palette = lcd.palette()
        # lcd.setPalette(QtGui.QPalette(Qt.red))
        if hasattr(Qt, color):
            palette.setBrush(palette.WindowText, getattr(Qt, color))
            palette.setColor(palette.Background, QtGui.QColor(0, 0, 0))
        lcd.setPalette(palette)


if __name__ == '__main__':
    app = QtWidgets.QApplication(sys.argv)

    form = QtWidgets.QWidget()
    win = QtWidgets.QMainWindow()
    win.setVisible(False)
    area = DockArea()
    win.setCentralWidget(area)
    win.setWindowTitle('Chrono Timer')

    # prog = ChronoTimer(area, dict(hours=1, minutes=0, seconds=0))
    prog = ChronoTimer(area)
    win.show()
    sys.exit(app.exec_())



from pathlib import Path

from pymodaq.utils.config import Config
from qtpy import QtWidgets

config = Config()


def select_file_filter(start_path=config('data_saving', 'h5file', 'save_path'), save=True, ext=None, filter=None,
                force_save_extension=False):
    """Opens a selection file popup for loading or saving a file

    Parameters
    ----------
    start_path: str or Path
        The strating point in the file/folder system to open the popup from
    save: bool
        if True, ask you to enter a filename (with or without extension)
    ext: str
        the extension string, e.g. xml, h5, png ...
    filter: list of string
        list of possible extensions, mostly valid for loading
    force_save_extension: bool
        if True force the extension of the saved file to be set to ext

    Returns
    -------
    Path: the Path object of the file to save or load
    str: the selected filter
    """
    if ext is None:
        ext = '.h5'

    if filter is None:
        if not save:
            if not isinstance(ext, list):
                ext = [ext]
            filter = "Data files ("
            for ext_tmp in ext:
                filter += '*.' + ext_tmp + " "
            filter += ")"

    if start_path is not None:
        if not isinstance(start_path, str):
            start_path = str(start_path)
    if save:
        fname, selected_filter = QtWidgets.QFileDialog.getSaveFileName(None, 'Enter a file name', start_path,
                                                                       filter)
    else:
        fname, selected_filter = QtWidgets.QFileDialog.getOpenFileName(None, 'Select a file name', start_path, filter)

    if fname != '':  # execute if the user didn't cancel the file selection
        fname = Path(fname)
        if save and force_save_extension:
            parent = fname.parent
            filename = fname.stem
            fname = parent.joinpath(filename + "." + ext)  # forcing the right extension on the filename
    return fname, selected_filter


def select_file(start_path=config('data_saving', 'h5file', 'save_path'), save=True, ext=None, filter=None,
                force_save_extension=False):
    """Opens a selection file popup for loading or saving a file

    Parameters
    ----------
    start_path: str or Path
        The strating point in the file/folder system to open the popup from
    save: bool
        if True, ask you to enter a filename (with or without extension)
    ext: str
        the extension string, e.g. xml, h5, png ...
    filter: list of string
        list of possible extensions, mostly valid for loading
    force_save_extension: bool
        if True force the extension of the saved file to be set to ext

    Returns
    -------
    Path: the Path object of the file to save or load
    """
    fname, selected_filter = select_file_filter(start_path, save, ext, filter,
                                                force_save_extension)
    return fname  # fname is a Path object


if __name__ == '__main__':
    import sys
    app = QtWidgets.QApplication(sys.argv)
    file = select_file(save=True, filter="Images (*.png *.xpm *.jpg);;Text files (*.txt);;XML files (*.xml)")
    print(file)
    sys.exit(app.exec_())



import numpy as np
from pyqtgraph.parametertree import Parameter
from qtpy import QtCore, QtWidgets, QtGui
from qtpy.QtCore import QPointF, Slot, Signal, QObject
from typing import List, Tuple

from pyqtgraph import LinearRegionItem

from pymodaq.utils import data as data_mod
from pymodaq.utils import daq_utils as utils
from pymodaq.utils import math_utils as mutils
from pymodaq.utils.managers.roi_manager import ROIManager, LinearROI, RectROI
from pymodaq.utils.plotting.items.crosshair import Crosshair
from pymodaq.utils.plotting.items.image import UniformImageItem
from pymodaq.utils.plotting.data_viewers.viewer1Dbasic import Viewer1DBasic
from pymodaq.utils.logger import set_logger, get_module_name


from pymodaq.post_treatment.process_to_scalar import DataProcessorFactory


logger = set_logger(get_module_name(__file__))


data_processors = DataProcessorFactory()


class Filter:

    def __init__(self):
        self._is_active = False
        self._slot_to_send_data = None

    def register_activation_signal(self, activation_signal):
        activation_signal.connect(lambda x: self.set_active(x))

    def register_target_slot(self, slot):
        self._slot_to_send_data = slot

    @Slot(bool)
    def set_active(self, activate=True):
        self._is_active = activate

    def filter_data(self, data: data_mod.DataFromPlugins):
        if self._is_active:
            filtered_data = self._filter_data(data)
            if filtered_data is not None and self._slot_to_send_data is not None:
                self._slot_to_send_data(filtered_data)

    def _filter_data(self, data: data_mod.DataFromPlugins):
        raise NotImplementedError


class Filter1DFromCrosshair(Filter):
    def __init__(self, crosshair: Crosshair):
        """
        Extract data along a crosshair using coordinates and data displayed in graph_items such as  imageItems
        Parameters
        ----------
        crosshair : Crosshair
        """
        super().__init__()
        self.crosshair = crosshair
        self._x, self._y = 0., 0.
        self._axis: data_mod.Axis = None

    def update_axis(self, axis: data_mod.Axis):
        self._axis = axis

    def _filter_data(self, data: data_mod.DataFromPlugins):
        data_dict = dict([])
        if data is not None:
            axis = data.get_axis_from_index(0, create=False)[0]
            if axis is not None:
                self.update_axis(axis)

                self._x, self._y = self.crosshair.get_positions()
                ind_x = self._axis.find_index(self._x)
                for label, dat in zip(data.labels, data.data):
                    data_dict[label] = dict(pos=self._axis.get_data()[ind_x], value=dat[ind_x])
        return data_dict


class Filter2DFromCrosshair(Filter):
    def __init__(self, crosshair: Crosshair, graph_items, image_keys):
        """
        Extract data along a crosshair using coordinates and data displayed in graph_items such as  imageItems
        Parameters
        ----------
        crosshair : (Crosshair)
        graph_items : (dict)
        image_keys : (list) list of string identifier to link datas to their graph_items. This means that in
            _filter_data, datas['data'][key] is plotted on graph_items[key] for key in image_keys
        """
        super().__init__()
        self._graph_items = graph_items
        self._image_keys = image_keys
        self.crosshair = crosshair
        self._x, self._y = 0., 0.

    def set_graph_items(self, graph_items):
        self._graph_items = graph_items

    @Slot(bool)
    def set_active(self, activate=True):
        self._is_active = activate
        if activate:
            self.crosshair.crosshair_dragged.emit(*self.crosshair.get_positions())

    def _filter_data(self, datas: data_mod.DataFromPlugins):
        data_dict = dict([])
        if datas is not None:
            self._x, self._y = self.crosshair.get_positions()
            data_type = datas.distribution
            for data_index in range(len(self._image_keys)):
                if data_index < len(datas.data):
                    data = datas.data[data_index]
                    image_type = self._image_keys[data_index]
                    if data_type == 'uniform':
                        data_dict[image_type] = self.get_data_from_uniform(image_type, data)
                    elif data_type == 'spread':
                        data_dict[image_type] = self.get_data_from_spread(image_type, data)
        return data_dict

    def get_data_from_uniform(self, data_key, data):
        hor_axis, ver_axis = \
            np.linspace(0, self._graph_items[data_key].width() - 1, self._graph_items[data_key].width()),\
            np.linspace(0, self._graph_items[data_key].height() - 1, self._graph_items[data_key].height())

        indx, indy = self.mapfromview(self._x, self._y, data_key)

        data_H_index = slice(None, None, 1)
        data_V_index = slice(None, None, 1)
        H_indexes = (utils.rint(indy), data_H_index)
        V_indexes = (data_V_index, utils.rint(indx))

        out_of_bounds = False
        if 0 <= H_indexes[0] < len(ver_axis):
            hor_data = data[H_indexes]
        else:
            out_of_bounds = True
            hor_data = np.zeros(hor_axis.shape)
        if 0 <= V_indexes[1] < len(hor_axis):
            ver_data = data[V_indexes]
        else:
            out_of_bounds = True
            ver_data = np.zeros(ver_axis.shape)
        if out_of_bounds:
            ind_data = 0.
        else:
            ind_data = data[utils.rint(indy), utils.rint(indx)]
        return LineoutData(hor_axis=hor_axis, ver_axis=ver_axis, hor_data=hor_data, ver_data=ver_data,
                           int_data=ind_data)

    def get_data_from_spread(self, data_key, data):
        data_H_index = slice(None, None, 1)
        data_V_index = slice(None, None, 1)
        posx, posy = self.mapfromview(self._x, self._y, data_key)

        points, data = self._graph_items[data_key].get_points_at(axis='y', val=posy)
        x_sorted_indexes = np.argsort(points[:, 0])
        hor_axis = points[x_sorted_indexes, 0][data_H_index]

        hor_data = data[x_sorted_indexes][data_H_index]

        points, data = self._graph_items[data_key].get_points_at(axis='x', val=posx)
        y_sorted_indexes = np.argsort(points[:, 1])
        ver_axis = points[y_sorted_indexes, 1][data_V_index]

        ver_data = data[y_sorted_indexes][data_V_index]

        return LineoutData(hor_axis=hor_axis, ver_axis=ver_axis, hor_data=hor_data, ver_data=ver_data,
                           int_data=self._graph_items[data_key].get_val_at((posx, posy)))

    def mapfromview(self, x, y, item_key='red'):
        """
        get item coordinates from view coordinates
        Parameters
        ----------
        x: (float) x coordinate in the view reference frame
        y: (float) y coordinate in the view refernece frame

        Returns
        -------
        x: (float) coordinate in the item reference frame
        y: (float) coordinate in the item reference frame
        """
        point = self._graph_items[item_key].mapFromView(QPointF(x, y))
        return point.x(), point.y()


class Filter1DFromRois(Filter):
    """

    Parameters
    ----------
    roi_manager:ROIManager
    graph_item: PlotItems
    """
    def __init__(self, roi_manager: ROIManager):

        super().__init__()
        self._roi_settings = roi_manager.settings
        self._ROIs = roi_manager.ROIs
        self._axis: data_mod.Axis = None

    def update_axis(self, axis: data_mod.Axis):
        self._axis = axis

    def _filter_data(self, data: data_mod.DataFromPlugins) -> dict:
        data_dict = dict([])
        try:
            axis = data.get_axis_from_index(0, create=False)[0]
            if axis is not None:
                self.update_axis(axis)
            if data is not None:
                for roi_key, roi in self._ROIs.items():
                    try:
                        data_index = data.labels.index(self._roi_settings['ROIs', roi_key, 'use_channel'])
                    except ValueError:
                        data_index = 0
                    data_dict[roi_key] = self.get_data_from_roi(roi, self._roi_settings.child('ROIs', roi_key),
                                                                data, data_index)
        except Exception as e:
            pass
        return data_dict

    def get_data_from_roi(self, roi: LinearROI,  roi_param: Parameter, data: data_mod.DataWithAxes, data_index=0):
        if data is not None:
            _slice = self.get_slice_from_roi(roi, data)
            sub_data = data.isig[_slice]
            if sub_data.size != 0:
                processed_data = data_processors.get(roi_param['math_function']).process(sub_data)
            else:
                processed_data = None
            if processed_data is None:
                return LineoutData()
            else:
                if len(sub_data.axes) == 0:
                    pass
                return LineoutData(hor_axis=sub_data.axes[0], hor_data=sub_data.data[data_index],
                                   int_data=processed_data.data[data_index])

    def get_slice_from_roi(self, roi: RectROI, data: data_mod.DataWithAxes) -> slice:
        ind_x_min, ind_x_max = data.get_axis_from_index(data.sig_indexes[0])[0].find_indexes(roi.getRegion())
        size = data.get_axis_from_index(0)[0].size
        ind_x_min = int(min(max(ind_x_min, 0), size))
        ind_x_max = int(max(0, min(ind_x_max, size)))
        return slice(ind_x_min, ind_x_max)


class Filter2DFromRois(Filter):
    """Filters 2D data using 2D ROIs

    Parameters
    ----------
    roi_manager: ROIManager
    graph_item: UniformImageItem or SpreadImageItem
        The graphical item where data and ROIs are plotted
    image_keys : (list) list of string identifier to link datas to their graph_items. This means that in
        _filter_data, datas.data[key] is plotted on graph_items[key] for key in image_keys
    """
    def __init__(self, roi_manager: ROIManager, graph_item: UniformImageItem, image_keys):

        super().__init__()
        self._roi_settings = roi_manager.settings
        self._image_keys = image_keys
        self._graph_item = graph_item
        self.axes = (0, 1)
        self._ROIs = roi_manager.ROIs

    def _filter_data(self, data: data_mod.DataFromPlugins) -> dict:
        data_dict = dict([])
        try:
            if data is not None:
                for roi_key, roi in self._ROIs.items():
                    image_key = self._roi_settings['ROIs', roi_key, 'use_channel']
                    image_index = self._image_keys.index(image_key)

                    sub_data = data.deepcopy()
                    sub_data.data = [data[image_index]]
                    data_dict[roi_key] = self.get_xydata_from_roi(roi, sub_data,
                                                                  self._roi_settings['ROIs', roi_key, 'math_function'])
        except Exception as e:
            pass
        return data_dict

    def get_slices_from_roi(self, roi: RectROI, data: data_mod.DataWithAxes) -> Tuple[slice]:
        x, y = roi.pos().x(), roi.pos().y()
        width, height = roi.size().x(), roi.size().y()
        size_x = data.get_axis_from_index(1)[0].size
        size_y = data.get_axis_from_index(0)[0].size
        ind_x_min = int(min(max(x, 0), size_x))
        ind_y_min = int(min(max(y, 0), size_y))
        ind_x_max = int(max(0, min(x+width, size_x)))
        ind_y_max = int(max(0, min(y+height, size_y)))

        return slice(ind_y_min,ind_y_max), slice(ind_x_min, ind_x_max)

    def get_xydata_from_roi(self, roi, data: data_mod.DataWithAxes, math_function: str):

        if data is not None:
            if data.distribution.name == 'spread':
                xvals, yvals, data = self.get_xydata_spread(data, roi)
                ind_xaxis = np.argsort(xvals)
                ind_yaxis = np.argsort(yvals)
                xvals = xvals[ind_xaxis]
                yvals = yvals[ind_yaxis]
                data_H = data[ind_xaxis]
                data_V = data[ind_yaxis]
                int_data = np.array([np.mean(data)])
                math_data = int_data
            else:
                xvals, yvals, data_array = self.get_xydata(data.data[0], roi)
                slices = self.get_slices_from_roi(roi, data)
                sub_data = data.isig[slices[0], slices[1]]
                data_H = np.mean(data_array, axis=0)
                data_V = np.mean(data_array, axis=1)
                int_data = np.array([np.mean(data_array)])
                math_data = data_processors.get(math_function).process(sub_data).data

            return LineoutData(hor_axis=xvals, ver_axis=yvals, hor_data=data_H, ver_data=data_V, int_data=int_data,
                               math_data=math_data)

    def get_xydata(self, data: np.ndarray, roi: RectROI):
        data, coords = self.data_from_roi(data, roi)

        if data is not None:
            xvals = np.linspace(np.min(np.min(coords[1, :, :])), np.max(np.max(coords[1, :, :])),
                                data.shape[1])
            yvals = np.linspace(np.min(np.min(coords[0, :, :])), np.max(np.max(coords[0, :, :])),
                                data.shape[0])
        else:
            xvals = yvals = data = np.array([])
        return xvals, yvals, data

    def data_from_roi(self, data, roi):
        data, coords = roi.getArrayRegion(data, self._graph_item, self.axes, returnMappedCoords=True)
        return data, coords

    def get_xydata_spread(self, data, roi):
        xvals = []
        yvals = []
        data_out = []
        for ind in range(data.shape[0]):
            # invoke the QPainterpath of the ROI (from the shape method)
            if roi.shape().contains(QPointF(data.get_axis_from_index(0)[0].get_data()[ind] - roi.pos().x(),
                                            data.get_axis_from_index(0)[1].get_data()[ind] - roi.pos().y())):
                xvals.append(data.get_axis_from_index(0)[0].get_data()[ind])
                yvals.append(data.get_axis_from_index(0)[1].get_data()[ind])
                data_out.append(data[0][ind])
        data_out = np.array(data_out)
        xvals = np.array(xvals)
        yvals = np.array(yvals)
        return xvals, yvals, data_out


class LineoutData:
    def __init__(self, hor_axis=np.array([]), ver_axis=np.array([]), hor_data=np.array([]), ver_data=np.array([]),
                 int_data: np.ndarray = None, math_data: List[np.ndarray] = None):
        super().__init__()
        if len(hor_axis) != len(hor_data):
            raise ValueError(f'Horizontal lineout data and axis must have the same size')
        if len(ver_axis) != len(ver_data):
            raise ValueError(f'Horizontal lineout data and axis must have the same size')

        self.hor_axis = hor_axis
        self.ver_axis = ver_axis
        self.hor_data = hor_data
        self.ver_data = ver_data
        if int_data is None:
            self.int_data = np.array([np.sum(self.ver_data)])
        else:
            self.int_data = int_data
        if math_data is None:
            math_data = self.int_data
        self.math_data = math_data


class FourierFilterer(QObject):
    filter_changed = Signal(dict)

    def __init__(self, parent=None):
        super().__init__()
        if parent is None:
            parent = QtWidgets.QWidget()

        self.parent = parent

        self.raw_data = None
        self.data = None
        self.data_fft = None
        self.filter = None
        self.xaxis = None
        self.yaxis = None
        self.xaxisft = None
        self.yaxisft = None

        self.frequency = 0
        self.phase = 0

        self.c = None
        self.viewer2D = None
        self.setUI()

    def setUI(self):
        self.vlayout = QtWidgets.QVBoxLayout()
        self.parent.setLayout(self.vlayout)

        form = QtWidgets.QWidget()
        self.viewer1D = Viewer1DBasic(form)
        self.vlayout.addWidget(form)
        self.fftbutton1D = QtWidgets.QPushButton()
        self.fftbutton1D.setText("")
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/icons/Icon_Library/FFT.png"), QtGui.QIcon.Normal,
                       QtGui.QIcon.Off)
        self.fftbutton1D.setIcon(icon)
        self.fftbutton1D.setCheckable(True)
        self.fftbutton1D.clicked.connect(self.update_plot)

        vbox = self.viewer1D.parent.layout()
        widg = QtWidgets.QWidget()
        hbox = QtWidgets.QHBoxLayout()
        widg.setLayout(hbox)
        vbox.insertWidget(0, widg)
        hbox.addWidget(self.fftbutton1D)
        hbox.addStretch()

        self.viewer1D.ROI = LinearRegionItem(values=[0, 100])
        self.viewer1D.plotwidget.plotItem.addItem(self.viewer1D.ROI)
        self.data_filtered_plot = self.viewer1D.plotwidget.plotItem.plot()
        self.data_filtered_plot.setPen('w')
        self.viewer1D.ROI.sigRegionChangeFinished.connect(self.set_data)

        self.viewer1D.ROIfft = LinearRegionItem()
        self.viewer1D.plotwidget.plotItem.addItem(self.viewer1D.ROIfft)
        self.viewer1D.ROIfft.sigRegionChangeFinished.connect(self.update_filter)

        self.parent.show()

    def calculate_fft(self):

        ftaxis, axis = mutils.ftAxis_time(len(self.xaxis), np.max(self.xaxis) - np.min(self.xaxis))
        self.xaxisft = ftaxis / (2 * np.pi)
        self.data_fft = mutils.ft(self.data)

    def show_data(self, data):
        """
        show data and fft
        Parameters
        ----------
        data: (dict) with keys 'data', optionally 'xaxis' and 'yaxis'
        """
        try:
            self.raw_data = data

            if 'xaxis' in data:
                self.xaxis = data['xaxis']
            else:
                self.xaxis = np.arange(0, data['data'].shape[0], 1)
                self.raw_data['xaxis'] = self.xaxis
            # self.viewer1D.ROI.setRegion((np.min(self.xaxis), np.max(self.xaxis)))
            self.set_data()
        except Exception as e:
            logger.exception(str(e))

    def set_data(self):
        xlimits = self.viewer1D.ROI.getRegion()
        indexes = mutils.find_index(self.raw_data['xaxis'], xlimits)
        self.data = self.raw_data['data'][indexes[0][0]:indexes[1][0]]
        self.xaxis = self.raw_data['xaxis'][indexes[0][0]:indexes[1][0]]
        try:
            self.calculate_fft()
        except Exception as e:
            logger.exception(str(e))
        self.viewer1D.x_axis = self.xaxis
        self.update_plot()

    def update_filter(self):
        try:
            xmin, xmax = self.viewer1D.ROIfft.getRegion()
            self.filter = mutils.gauss1D(self.xaxisft, np.mean([xmin, xmax]), xmax - xmin)
            self.data = np.real(mutils.ift(self.filter * self.data_fft))
            index = np.argmax(self.filter * self.data_fft)
            self.frequency = self.xaxisft[index]
            self.phase = np.angle(self.data_fft[index])

            self.filter_changed.emit(dict(frequency=self.frequency, phase=self.phase))
            self.update_plot()
        except Exception as e:
            logger.exception(str(e))

    def update_plot(self):

        if self.fftbutton1D.isChecked():
            if self.data_fft is not None:
                if self.filter is not None:
                    self.viewer1D.show_data([np.abs(self.data_fft), np.max(np.abs(self.data_fft)) * self.filter])
                else:
                    self.viewer1D.show_data([np.abs(self.data_fft)])
                self.viewer1D.x_axis = dict(data=self.xaxisft, label='freq.')
                self.viewer1D.ROIfft.setVisible(True)
                self.viewer1D.ROI.setVisible(False)
                self.data_filtered_plot.setVisible(False)
        else:
            if self.raw_data is not None:
                self.viewer1D.show_data([self.raw_data['data']])
                self.viewer1D.x_axis = dict(data=self.raw_data['xaxis'], label='Pxls')
                self.data_filtered_plot.setData(self.xaxis, self.data)
                self.data_filtered_plot.setVisible(True)
                self.viewer1D.ROIfft.setVisible(False)
                self.viewer1D.ROI.setVisible(True)


if __name__ == '__main__':
    import sys

    app = QtWidgets.QApplication(sys.argv)
    prog = FourierFilterer()

    from pymodaq.utils.daq_utils import gauss1D

    xdata = np.linspace(0, 400, 401)
    x0 = 50
    dx = 20
    tau = 27
    tau2 = 100
    ydata_gauss = 10 * gauss1D(xdata, x0, dx) + np.random.rand(len(xdata))
    ydata_expodec = np.zeros((len(xdata)))
    ydata_expodec[:50] = 10 * gauss1D(xdata[:50], x0, dx, 2)
    ydata_expodec[50:] = 10 * np.exp(-(xdata[50:] - x0) / tau)  # +10*np.exp(-(xdata[50:]-x0)/tau2)
    ydata_expodec += 2 * np.random.rand(len(xdata))
    ydata_sin = 10 + 2 * np.sin(2 * np.pi * 0.1 * xdata - np.deg2rad(55)) + np.sin(
        2 * np.pi * 0.008 * xdata - np.deg2rad(-10)) + 2 * np.random.rand(len(xdata))

    prog.show_data(dict(data=ydata_sin, xaxis=xdata))
    sys.exit(app.exec_())



import numpy as np

from pymodaq.utils.logger import set_logger, get_module_name
from pymodaq.utils.gui_utils.custom_app import CustomApp
from pymodaq.utils.gui_utils.dock import DockArea, Dock
from pymodaq.utils.config import Config, get_set_preset_path
from qtpy import QtWidgets
from qtpy.QtCore import Signal, QLocale

from pymodaq.utils import gui_utils as gutils
from pymodaq.utils import daq_utils as utils


config = Config()
logger = set_logger(get_module_name(__file__))


class NonLinearScanner(CustomApp):

    positions_signal = Signal(np.ndarray)

    # list of dicts enabling the settings tree on the user interface
    params = [
        {'title': 'Init Position:', 'name': 'ini_pos', 'type': 'float', 'value': 0, },
        {'title': 'Saturation pos:', 'name': 'satu_pos', 'type': 'float', 'value': 5, },
        {'title': 'Start scan:', 'name': 'start_scan', 'type': 'float', 'value': 4, },
        {'title': 'Stop Scan:', 'name': 'stop_scan', 'type': 'float', 'value': 3, },
        {'title': 'Step Scan:', 'name': 'step_scan', 'type': 'float', 'value': -0.05, },
        {'title': 'Apply', 'name': 'apply', 'type': 'action',},

    ]

    def __init__(self, dockarea):
        
        super().__init__(dockarea)
        self.setup_ui()

    def setup_actions(self):
        '''
        subclass method from ActionManager
        '''
        logger.debug('setting actions')

        logger.debug('actions set')

    def setup_docks(self):
        '''
        subclass method from CustomApp
        '''
        logger.debug('setting docks')
        self.dock_settings = Dock('Settings', size=(350, 350))
        self.dockarea.addDock(self.dock_settings, 'left')
        self.dock_settings.addWidget(self.settings_tree, 10)
        logger.debug('docks are set')

    def connect_things(self):
        '''
        subclass method from CustomApp
        '''
        logger.debug('connecting things')
        self.settings.child('apply').sigActivated.connect(self.emit_positions)
        logger.debug('connecting done')

    def setup_menu(self):
        '''
        subclass method from CustomApp
        '''
        logger.debug('settings menu')


        logger.debug('menu set')

    def value_changed(self, param):
        logger.debug(f'calling value_changed with param {param.name()}')

        logger.debug(f'Value change applied')

    def update_positions(self):
        xini = self.settings.child('ini_pos').value()
        sat = self.settings.child('satu_pos').value()
        start = self.settings.child('start_scan').value()
        stop = self.settings.child('stop_scan').value()
        step = self.settings.child('step_scan').value()

        return np.concatenate((np.array([xini, sat]), utils.linspace_step(start, stop, step), np.array([xini])))

    def emit_positions(self):
        self.positions_signal.emit(np.transpose(np.array([self.update_positions()])))


def main():
    import sys
    from pathlib import Path
    app = QtWidgets.QApplication(sys.argv)

    from pymodaq.dashboard import DashBoard

    win = QtWidgets.QMainWindow()
    area = DockArea()
    win.setCentralWidget(area)
    win.resize(1000, 500)
    win.setWindowTitle('PyMoDAQ Dashboard')

    dash = DashBoard(area)
    file = Path(get_set_preset_path()).joinpath(f"{config('presets', 'default_preset_for_scan')}.xml")
    if file.exists():
        dash.set_preset_mode(file)
        dash.load_scan_module()


    mainwindow = QtWidgets.QMainWindow()
    dockarea = DockArea()
    mainwindow.setCentralWidget(dockarea)

    prog = NonLinearScanner(dockarea)

    QtWidgets.QApplication.processEvents()
    prog.positions_signal.connect(dash.scan_module.scanner.update_tabular_positions)

    mainwindow.show()
    sys.exit(app.exec_())


if __name__ == '__main__':
    main()



import warnings
from qtpy import QtWidgets, QtCore


def deprecation_msg(message, stacklevel=2):
    warnings.warn(message, DeprecationWarning, stacklevel=stacklevel)


def user_warning(message, stacklevel=3):
    warnings.warn(message, UserWarning, stacklevel=stacklevel)


MESSAGE_SEVERITIES = ['critical', 'information', 'question', 'warning']


def messagebox(severity='warning', title='this is a title', text='blabla'):
    """
    Display a popup messagebox with a given severity
    Parameters
    ----------
    severity: (str) one in ['critical', 'information', 'question', 'warning']
    title: (str) the displayed popup title
    text: (str) the displayed text in the message

    Returns
    -------
    bool: True if the user clicks on Ok
    """
    assert severity in MESSAGE_SEVERITIES
    messbox = getattr(QtWidgets.QMessageBox, severity)
    ret = messbox(None, title, text)
    return ret == QtWidgets.QMessageBox.Ok


def dialog(title='', message='', widget=None):
    dlg = QtWidgets.QDialog()
    dlg.setWindowTitle(title)
    dlg.setLayout(QtWidgets.QVBoxLayout())
    label = QtWidgets.QLabel(message)
    label.setAlignment(QtCore.Qt.AlignCenter)
    dlg.layout().addWidget(label)
    dlg.layout().addWidget(widget)
    button_box = QtWidgets.QDialogButtonBox(QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel)
    dlg.layout().addWidget(button_box)
    button_box.accepted.connect(dlg.accept)
    button_box.rejected.connect(dlg.reject)
    ret = dlg.exec()
    return ret


def show_message(message="blabla", title="Error"):
    DeprecationWarning('Using show_message is deprecated, please use messagebox in daq_utils.messenger module')
    messagebox('Warning', title=title, text=message)


if __name__ == '__main__':
    import sys
    from pymodaq.utils.plotting.data_viewers.viewer1D import Viewer1D
    app = QtWidgets.QApplication([])
    widget = QtWidgets.QWidget()
    viewer = Viewer1D(widget)
    ret = dialog('my dialog', 'this is a dialog do you want to proceed', widget)
    print(ret)
    sys.exit(app.exec_())





# -*- coding: utf-8 -*-
"""
Created the 15/11/2022

@author: Sebastien Weber
"""

import logging
import numpy as np

from pymodaq.utils.logger import set_logger, get_module_name
from pymodaq.utils.config import Config
from pymodaq.utils.managers.modules_manager import ModulesManager
from pymodaq.utils.h5modules import module_saving, data_saving

from pymodaq.utils.abstract.logger import AbstractLogger
from pymodaq.utils import daq_utils as utils
from .saving import H5Saver
from pymodaq.utils.data import DataToExport


config = Config()

logger = set_logger(get_module_name(__file__))


class H5LogHandler(logging.StreamHandler):
    def __init__(self, h5saver):
        super().__init__()
        self.h5saver = h5saver
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        self.setFormatter(formatter)

    def emit(self, record):
        msg = self.format(record)
        self.h5saver.add_log(msg)


class H5Logger(AbstractLogger):
    def __init__(self, modules_manager, *args, **kwargs):

        self.title = self.__class__.__name__

        self.modules_manager: ModulesManager = modules_manager
        self.h5saver = H5Saver(*args, save_type='logger', **kwargs)

        self.module_and_data_saver = module_saving.LoggerSaver(self)
        for det in self.modules_manager.detectors_all:
            det.module_and_data_saver = module_saving.DetectorEnlargeableSaver(det)
        self.module_and_data_saver.h5saver = self.h5saver  # will update its h5saver and all submodules's h5saver

    def close(self):
        self.h5saver.close_file()

    @property
    def settings_tree(self):
        return self.h5saver.settings_tree

    @property
    def settings(self):
        return self.h5saver.settings

    def init_logger(self, settings):
        self.h5saver.init_file(update_h5=True, metadata=dict(settings=settings))
        self.h5saver.flush()
        self.module_and_data_saver.h5saver = self.h5saver
        logger_node = self.module_and_data_saver.get_set_node(new=True)
        return True

    def get_handler(self):
        return H5LogHandler(self.h5saver)

    def add_detector(self, name, settings):
        pass

    def add_actuator(self, name, settings):
        pass

    def add_data(self, dte: DataToExport):
        self.module_and_data_saver.add_data(dte)

        self.settings.child('N_saved').setValue(self.settings.child('N_saved').value() + 1)

    def stop_logger(self):
        self.h5saver.flush()
